"""Backends for labrad connections

Provides a backend connection that underlies the wrapper client object.
"""

from __future__ import absolute_import

import threading

from concurrent.futures import Future
from twisted.internet import defer, reactor

from labrad import concurrent, constants as C, support, thread, types as T
from labrad.wrappers import getConnection


class TwistedConnection(object):
    def __init__(self, cxn, spawn_kw=None):
        """Create a synchronous wrapper around an asynchronous connection.

        Args:
            cxn (labrad.protocol.LabradProtocol): The asynchronous protocol
                instance for which to provide a synchronous interface.
            spawn_kw (dict): Keyword arguments to pass to labrad.backend.connect
                to spawn a new connection. If None, spawning will fail.
        """
        self.cxn = cxn
        self.name = cxn.name
        self.ID = cxn.ID
        self.host = cxn.host
        self.port = cxn.port
        self._spawn_kw = spawn_kw
        self._connected = threading.Event()
        self._connected.set()

        # Setup a coroutine that will clear the connected flag when the
        # connection is lost. We launch this but do not wait for the result of
        # the future because we want this to happen asynchronously in the
        # background.
        @defer.inlineCallbacks
        def handle_disconnect():
            try:
                yield cxn.onDisconnect()
            except Exception:
                pass
            self._connected.clear()
        concurrent.call_future(handle_disconnect)

    @property
    def connected(self):
        return self._connected.is_set()

    def disconnect(self):
        if self.connected:
            concurrent.call_future(self.cxn.disconnect).result()

    def spawn(self):
        """Start a new independent backend connection to the same manager."""
        if self._spawn_kw is None:
            raise ValueError("Cannot spawn from {}".format(self))
        return connect(**self._spawn_kw)

    def context(self):
        """Create a new context for use with this connection"""
        return concurrent.call_future(self.cxn.context).result()

    def sendRequest(self, target, records, *args, **kw):
        return concurrent.call_future(self.cxn.sendRequest, target, records,
                                      *args, **kw)

    def sendMessage(self, target, records, *args, **kw):
        return concurrent.call_future(self.cxn.sendMessage, target, records,
                                      *args, **kw).result()


def connect(host=C.MANAGER_HOST, port=None, name=None, timeout=C.TIMEOUT, **kw):
    """Create a backend connection to labrad.

    This connects to labrad asynchronously and then wraps the underlying async
    connection object in a synchronous TwistedConnection interface.
    """
    name = name or 'Python Client ({})'.format(support.getNodeName())

    thread.startReactor()
    future = concurrent.call_future(getConnection, host, port, name, **kw)
    cxn = future.result(timeout=timeout)

    # Make a dict of all params we were called with so that identical new
    # connections can be spawned from this one.
    spawn_kw = dict(host=host, port=port, name=name, timeout=timeout, **kw)

    return TwistedConnection(cxn, spawn_kw=spawn_kw)


class ManagerService:
    """Wraps a backend connection to provide a basic synchronous interface to the Manager."""
    def __init__(self, cxn):
        self.cxn = cxn

    def getServersList(self):
        """Get list of connected servers."""
        return self._getIDList(T.flatten(C.SERVERS_LIST, 'w'))

    def getServerInfo(self, serverID):
        """Get information about a server."""
        packet = [(C.HELP, T.flatten(serverID, 'w')),
                  (C.SETTINGS_LIST, T.flatten(serverID, 'w'))]
        resp = self._send(packet)
        descr, notes = resp[0][1]
        settings = self._reorderIDList(resp[1][1])
        return (descr, notes, settings)

    def getSettingsList(self, serverID):
        """Get list of settings for a server."""
        return self._getIDList(C.SETTINGS_LIST, T.flatten(serverID, 'w'))

    def getSettingInfo(self, serverID, settingID):
        """Get information about a setting."""
        packet = [(C.HELP, T.flatten((serverID, settingID), 'ww'))]
        resp = self._send(packet)
        description, accepts, returns, notes = resp[0][1]
        return (description, accepts, returns, notes)

    def getSettingInfoByName(self, serverID, settingName):
        """Get information about a setting using its name."""
        packet = [(C.HELP, T.flatten((serverID, settingName), 'ws')),
                  (C.LOOKUP, T.flatten((serverID, settingName), 'ws'))]
        resp = self._send(packet)
        description, accepts, returns, notes = resp[0][1]
        ID = resp[1][1][1]
        return (description, accepts, returns, notes, ID)

    def _send(self, packet, *args, **kw):
        """Send a request to the manager and wait for the result."""
        return self.cxn.sendRequest(C.MANAGER_ID, packet, *args, **kw).result()

    def _getIDList(self, setting, data=None):
        resp = self._send([(setting, data)])
        names = self._reorderIDList(resp[0][1])
        return names

    def _reorderIDList(self, L):
        return [(name, ID) for ID, name in L]





from labrad import types as T

HEADER_TYPE = T.parseTypeTag('(ww)iww')
PACKET_TYPE = T.parseTypeTag('(ww)iws')
RECORD_TYPE = T.parseTypeTag('wss')

def packetStream(packetHandler, endianness='>'):
    """A generator that assembles packets.

    Accepts a function packetHandler that will be called with four arguments
    whenever a packet is completed: source, context, request, records.
    """
    buf = ''
    while True:
        # get packet header (20 bytes)
        while len(buf) < 20:
            buf += yield 0
        hdr, buf = buf[:20], buf[20:]
        context, request, source, length = T.unflatten(hdr, HEADER_TYPE, endianness=endianness)

        # get packet data
        while len(buf) < length:
            buf += yield 0
        s, buf = buf[:length], buf[length:]

        # unflatten the data
        records = unflattenRecords(s, endianness=endianness)

        packetHandler(source, context, request, records)

def unflattenPacket(data, endianness='>'):
    """Unflatten a single labrad packet"""
    hdr, rest = data[:20], data[20:]
    context, request, source, length = T.unflatten(hdr, HEADER_TYPE, endianness)
    assert len(rest) == length
    records = unflattenRecords(rest, endianness)
    return context, request, source, records

def unflattenRecords(data, endianness='>'):
    """Unflatten a list of records from the data segment of a packet"""
    records = []
    s = T.Buffer(data)
    while len(s):
        ID, tag, data = T.unflatten(s, RECORD_TYPE, endianness)
        rec = ID, T.FlatData(data, T.parseTypeTag(tag), endianness)
        records.append(rec)
    return records

def flattenPacket(target, context, request, records, endianness='>'):
    """Flatten a packet to the specified target."""
    if isinstance(records, str):
        data = records
    else:
        kw = {'endianness': endianness}
        data = ''.join(flattenRecord(*rec, **kw) for rec in records)
    flat = PACKET_TYPE.flatten((context, request, target, data), endianness)
    return flat.bytes

def flattenRecords(records, endianness='>'):
    kw = {'endianness': endianness}
    return ''.join(flattenRecord(*rec, **kw) for rec in records)

def flattenRecord(ID, data, types=[], endianness='>'):
    """Flatten a piece of data into a record with datatype and property."""
    try:
        flat = T.flatten(data, types, endianness)
    except T.FlatteningError as e:
        e.msg = e.msg + "\nSetting ID %s." % (ID,)
        raise
    flat_record = RECORD_TYPE.flatten((ID, str(flat.tag), str(flat.bytes)),
                                      endianness)
    return flat_record.bytes




# Copyright (C) 2007  Matthew Neeley
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
labrad.decorators

Decorators that help in creating LabRAD servers.
"""

from __future__ import absolute_import

import functools
import inspect
import itertools
import types

import twisted.internet.defer as defer

from labrad import types as T, util

def _product(lists):
    """Return the cartesian product of a list of lists."""
    if not len(lists): return [[]]
    return [[h] + t for h in lists[0] for t in _product(lists[1:])]

def setting(lr_ID, lr_name=None, returns=[], unflatten=True, **params):
    """Decorator to turn a class method into a remotely accessible setting.

    This just creates a Setting class object which does the actual business.
    """
    def decorator(func):
        try:
            handler = Setting(func, lr_ID, lr_name, returns, unflatten, **params)
            func = handler.func  # might have gotten wrapped with inlineCallbacks
            func.ID = handler.ID
            func.name = handler.name
            func.accepts = handler.accepts
            func.returns = handler.returns
            func.handleRequest = handler.handleRequest
            func.getRegistrationInfo = handler.getRegistrationInfo
            return func
        except Exception:
            print 'Error in setting {} ({}):'.format(func.__name__, lr_ID)
            raise
    return decorator

class Setting(object):
    """Represents a single setting within a Server object.
    """
    def __init__(self, func, lr_ID, lr_name, returns, unflatten, **params):
        """Setting constructor
        Args:

            func (callable):   Function to implement the setting.
            lr_ID (int):       The ID number of the setting used in the labrad
                               protocol
            lr_name (str):     The setting name.  By default, this is derived
                               from the function name.
            returns:           The labrad type tag or list of tags the setting
                               returns
            unflatten (bool):  Request automatic unflattening of incoming data.
                               (default True)
            **params:          Additional keyword arguments indicate arguments
                               to the setting.  Each keyword should be a string
                               matching one of the formal parameters of func,
                               while the value is the type tag or list of tags
                               accepted by that parameter.

        If unflattening is requested, pylabrad will use the default unflattening
        for the data.  Otherwise, arguments will receive FlatData objects.  If
        there are multiple arguments, the top-level tuple will still be
        unpacked."""

        self.func = func
        self.ID = lr_ID
        self.name = lr_name or func.__name__
        self.returns = [returns] if isinstance(returns, basestring) else returns
        self.unflatten = unflatten
        self.description, self.notes = util.parseSettingDoc(func.__doc__)
        self.__doc__ = "Setting wrapper for {}\n\n{}".format(func.__name__, func.__doc__)
        
        ###
        # We need to validate the arguments.  Things we need to checks:
        #
        # 1) All parameters must match function arguments
        # 2) Function arguments with no specified parameter default to '?'
        # 3) The empty tag '' is only allowd on the first argument, and only
        #    if all other arguments are optional.
        # 4) If more than one argument are required, we expect to always receive
        #    a tuple and unpack it
        # 5) If only a single argument is allowed, we never unpack tuples
        # 6) If both =1 and >1 arguments are allowed, it is ambiguous whether to
        #    unpack tuples, so this case is not allowed:  The first argument
        #    cannot be a tuple or '?' tag if the second argument is optional.
        
        argspec = inspect.getargspec(self.func)
        args = argspec.args[2:] # Skip 'self' and context data arguments.

        if inspect.isgeneratorfunction(func):
            self.func = defer.inlineCallbacks(func)

        for arg in args:
            if arg not in params:
                params[arg] = ['?']

        for p in params.keys():
            if p not in args:
                raise ValueError("Setting parameter {} not accepted by function".format(p))
            if isinstance(params[p], basestring):
                params[p] = [params[p]]

        Nparams = len(args)
        Noptional = len(argspec.defaults) if argspec.defaults else 0
        Nrequired = Nparams - Noptional

        if Nrequired > 1:
            self.expand_cluster = "always"
        elif Nparams > 1:
            self.expand_cluster = "optional"
        else: # one or fewer arguments
            self.expand_cluster = "never"

        self.allow_none = Nrequired == 0

        if Nparams:
            for tag in params[args[0]]:
                tt = T.parseTypeTag(tag)
                if isinstance(tt, T.LRNone) and Nrequired > 1:
                    raise ValueError("First argument {} cannot accept '' "
                          "unless other arguments are optional".format(args[0]))
                if isinstance(tt, (T.LRAny, T.LRCluster)) and self.expand_cluster=="optional":
                    raise ValueError("First argument {} cannot accept type {} "
                          "because other arguments are optional".format(args[0], tt))
            for arg in args[1:]:
                for tag in params[arg]:
                    if isinstance(T.parseTypeTag(tag), T.LRNone):
                        raise ValueError("Argument {} cannot accept ''".format(arg))

        # Build list of accepted data types.
        # This is basically every combination of accepted types for each agrument,
        # including omitting any number of trailing optional arguments.

        accepted_types = []
        for i in range(Nrequired, Nparams+1):
            if i == 0:
                accepted_types.append('_')
            else:
                accept_tuples = itertools.product(*[params[arg] for arg in args[:i]])
                accepted_types.extend(combine_type_tags(x) for x in accept_tuples)
        self.accepts = accepted_types

    def handleRequest(self, server, c, flat_data):
        data_type = flat_data.tag

        if (self.expand_cluster == "always" or
            (self.expand_cluster == "optional"
             and isinstance(data_type, T.LRCluster))):
            if self.unflatten:
                data = flat_data.unflatten()
            else:
                data = flat_data.tag.partial_unflatten(flat_data.bytes,
                                                       flat_data.endianness)
            return self.func(server, c, *data)
        if self.allow_none and isinstance(data_type, T.LRNone):
            return self.func(server, c)

        if self.unflatten:
            data = flat_data.unflatten()
            return self.func(server, c, data)
        else:
            return self.func(server, c, flat_data)

    def getRegistrationInfo(self):
        return (self.ID, self.name, self.description,
                self.accepts, self.returns, self.notes)


def combine_type_tags(tags):
    """Combine one or more type tag strings into a single type tag.

    We cannot simply concatenate the tag strings because some of the input tags
    may have trailing comments.

    Args:
        tags (iterable[str]): The type tags to be combined into one type.

    Returns:
        (string): A tag representing a cluster of the given tags.
    """
    if len(tags) == 1:
        return tags[0]

    types = [T.parseTypeTag(tag) for tag in tags]
    return str(T.LRCluster(*types))


class MessageHandler(object):
    def __init__(self, func):
        self.func = func

    def handleMessage(self, c, data):
        return self.func(c, data)

    def handleNamedMessage(self, c, data):
        c.source, data = data
        return self.func(c, data)

def messageHandler(lr_ID, lr_name=None, returns=[], lr_num_params=2, **params):
    """Mark a server method as a remotely-accessible setting.

    The only required parameter is an integer ID.  You may
    also provide a name to override the name of the decorated
    function.  In addition, accepted types for each of the setting
    parameters may be provided as named parameters with a list of
    strings of allowed types.
    """
    def decorated(f):
        args, varargs, varkw, defaults = inspect.getargspec(f)
        args = args[lr_num_params:]

        # handle generators as defer.inlineCallbacks
        if inspect.isgeneratorfunction(f):
            f = defer.inlineCallbacks(f)

        # make sure that defined params are actually accepted by the function.
        # having extra params would not affect the running, but it is
        # unnecessary and hence may indicate other problems with the code
        for p in params:
            if p not in args:
                raise Exception("'%s' is not a valid parameter." % p)
            # turn single string annotations into lists
            if isinstance(params[p], str):
                params[p] = [params[p]]

        Nparams = len(args)
        Noptional = 0 if defaults is None else len(defaults)
        Nrequired = Nparams - Noptional

        if Nparams == 0:
            accepts_s = [''] # only accept notifier
            accepts_t = [T.parseTypeTag(s) for s in accepts_s]

            @functools.wraps(f)
            def handleRequest(self, c, data):
                return f(self, c)

        elif Nparams == 1:
            accepts_s = params.get(args[0], [])
            accepts_t = [T.parseTypeTag(s) for s in accepts_s]

            if Nrequired == 0:
                # if accepted types were specified, add '' to the list
                # we don't add '' if the list of accepted types is empty,
                # since this would make '' the ONLY accepted type
                if len(accepts_t) and T.LRNone() not in accepts_t:
                    accepts_s.append(': defaults [%s=%r]' \
                                     % (args[0], defaults[0]))
                    accepts_t.append(T.LRNone())

                @functools.wraps(f)
                def handleRequest(self, c, data):
                    if data is None:
                        return f(self, c)
                    return f(self, c, data)

            else:
                # nothing special to do here
                handleRequest = f

        else:
            # sanity checks to make sure that we'll be able to
            # correctly dispatch to the function when called
            if Nrequired <= 1:
                if args[0] not in params:
                    raise Exception('Must specify types for first argument '
                                    'when fewer than two args are required.')
                for s in params[args[0]]:
                    t = T.parseTypeTag(s)
                    if isinstance(t, (T.LRAny, T.LRCluster)):
                        raise Exception('Cannot accept cluster or ? in first '
                                        'arg when fewer than two args are '
                                        'required.')

            # '' is not allowed on first arg when Nrequired > 1
            types = [T.parseTypeTag(s) for s in params.get(args[0], [])]
            if Nrequired > 1 and T.LRNone() in types:
                raise Exception("'' not allowed when more than "
                                "one arg is required.")

            # '' is never allowed on args after the first.
            for p in args[1:]:
                types = [T.parseTypeTag(s) for s in params.get(p, [])]
                if T.LRNone() in types:
                    raise Exception("'' not allowed after first arg.")

            # allowed types are as follows:
            # one type for each parameter, with the number of
            # parameters ranging from the total number down to
            # and including the required number
            # we don't include any zero-length group
            groups = []
            for n in range(Nparams, Nrequired-1, -1):
                lists = [params.get(a, ['?']) for a in args[:n]]
                if len(lists):
                    groups += _product(lists)
                for i, group in reversed(list(enumerate(groups))):
                    # if there are any LRNones in the group, we remove it
                    ts = [T.parseTypeTag(t) for t in group]
                    if T.LRNone() in ts:
                        groups.pop(i)

            accepts_t = []
            accepts_s = []
            for group in groups:
                if len(group) > 1:
                    t = T.LRCluster(*[T.parseTypeTag(t) for t in group])
                    s = ', '.join('%s{%s}' % (sub_t, arg)
                                  for sub_t, arg in zip(t, args))
                    s = '(%s)' % s
                else:
                    t = T.parseTypeTag(group[0])
                    if isinstance(t, T.LRCluster):
                        raise Exception("Can't accept cluster in first param.")
                    s = '%s{%s}' % (group[0], args[0])
                # add information about default values of unused params
                if len(group) < Nparams:
                    defstr = ', '.join('%s=%r' % (args[n], defaults[n-Nrequired])
                                       for n in range(len(group), Nparams))
                    s = s + ': defaults [%s]' % defstr
                accepts_t.append(t)
                accepts_s.append(s)

            if Nrequired == 0:
                if T.LRNone() not in accepts_t:
                    defstr = ', '.join('%s=%r' % (a, d)
                                       for a, d in zip(args, defaults))
                    accepts_s.append(': defaults [%s]' % defstr)
                    accepts_t.append(T.LRNone())

                @functools.wraps(f)
                def handleRequest(self, c, data):
                    if isinstance(data, tuple):
                        return f(self, c, *data)
                    elif data is None:
                        return f(self, c)
                    else:
                        return f(self, c, data)
            else:
                @functools.wraps(f)
                def handleRequest(self, c, data):
                    if isinstance(data, tuple):
                        return f(self, c, *data)
                    else:
                        return f(self, c, data)

        f.ID = lr_ID
        f.name = lr_name or f.__name__
        f.accepts = accepts_s
        f.returns = returns
        f.handleRequest = handleRequest

        # this is the data that will be sent to the manager to
        # register this setting to be remotely callable
        f.description, f.notes = util.parseSettingDoc(f.__doc__)
        def getRegistrationInfo():
            return (f.ID, f.name, f.description,
                    f.accepts, f.returns, f.notes)
        f.getRegistrationInfo = getRegistrationInfo

        return f
    return decorated



# Copyright (C) 2007  Matthew Neeley
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
labrad.protocol

Defines the sending and receiving of packets on the network,
as well as the protocol for connecting to the Manager and
authenticating.
"""

import hashlib

from twisted.internet import reactor, protocol, defer
from twisted.internet.defer import inlineCallbacks, returnValue
from twisted.python import failure, log
import labrad.types as T

from labrad import auth, constants as C, crypto, errors, oauth, support, util
from labrad.stream import packetStream, flattenPacket


class LabradProtocol(protocol.Protocol):
    """Receive and send labrad packets."""

    def __init__(self):
        self.disconnected = False
        self._next_context = 1
        self._nextRequest = 1
        self.pool = set()
        self.requests = {}
        self.listeners = {}
        self._messageLock = defer.DeferredLock()
        self.clearCache()
        self.endianness = '>'
        self.request_handler = None
        # create a generator to assemble the packets
        self.packetStream = packetStream(self.packetReceived, self.endianness)
        self.packetStream.next() # start the packet stream

        self.onDisconnect = util.DeferredSignal()

    def set_address(self, host, port):
        """Store host and port of remote endpoint in protocol object.

        This is called by the connect function below after successfully
        establishing a connection. The host and port are used to cache our
        credentials after we authenticate successfully.

        Args:
            host (str): The hostname we are connected to.
            port (int): The TCP port number we are connected to.
        """
        self.host = host
        self.port = port

    def context(self):
        """Create a new communication context for this connection."""
        context = (0, self._next_context)
        self._next_context += 1
        return context

    # network events
    def connectionMade(self):
        # set the SO_KEEPALIVE option on all connections
        self.transport.setTcpNoDelay(True)
        self.transport.setTcpKeepAlive(True)

    def connectionLost(self, reason):
        """Called when the network connection is lost.

        This can be due to the disconnect() method being
        called, or because of some network error.
        """
        self.disconnected = True
        for d in self.requests.values():
            d.errback(Exception('Connection lost.'))
        if reason == protocol.connectionDone:
            self.onDisconnect.callback(None)
        else:
            self.onDisconnect.errback(reason)

    def disconnect(self):
        """Close down the connection to LabRAD."""
        self.disconnected = True
        self.transport.loseConnection()

    # sending
    def sendPacket(self, target, context, request, records):
        """Send a raw packet to the specified target."""
        raw = flattenPacket(target, context, request, records, endianness=self.endianness)
        self.transport.write(raw)

    @inlineCallbacks
    def sendMessage(self, target, records, context=(0, 0)):
        """Send a message to the specified target."""
        target, records = yield self._lookupNames(target, records)
        self.sendPacket(target, context, 0, records)

    @inlineCallbacks
    def sendRequest(self, target, records, context=(0, 0), timeout=None, unflatten=True):
        """Send a request to the given target server.

        Returns a deferred that will fire the resulting data packet when
        the request is completed, or will errback if the request times out
        or errors are returned from labrad.  The target server and settings
        may be given either as word IDs or string names.  If necessary,
        any string names will be looked up before the request is sent.
        Lookup results are cached to avoid lookup overhead on subsequent
        requests to the same server or settings.
        """
        target, records = yield self._lookupNames(target, records)
        resp = yield self._sendRequestNoLookup(target, records, context, timeout, unflatten)
        returnValue(resp)

    @inlineCallbacks
    def _lookupNames(self, server, records):
        """Translate server and setting names into IDs.

        We first attempt to look up these names in the local cache.
        If any are not found there, we fire off a request to the
        Manager to lookup the necessary IDs, and then cache the
        result.
        """
        records = list(records)

        # try to lookup server in cache
        if isinstance(server, str) and server in self._serverCache:
            server = self._serverCache[server]

        # try to lookup settings in cache
        if server in self._settingCache:
            settings = self._settingCache[server]
            for i, rec in enumerate(records):
                name = rec[0]
                if isinstance(name, str) and name in settings:
                    records[i] = (settings[name],) + tuple(rec[1:])

        # check to see whether there is still anything to look up
        settingLookups = [(i, rec[0]) for i, rec in enumerate(records)
                                      if isinstance(rec[0], str)]
        if isinstance(server, str) or len(settingLookups):
            # need to do additional lookup here
            if len(settingLookups):
                indices, names = zip(*settingLookups)
            else:
                indices, names = [], []
            # send the actual lookup request
            recs = [(C.LOOKUP, (server, names), ['w*s', 's*s'])]
            resp = yield self._sendRequestNoLookup(C.MANAGER_ID, recs)
            serverID, IDs = resp[0][1]
            # cache the results
            if isinstance(server, str):
                self._serverCache[server] = serverID
            server = serverID
            settings = self._settingCache.setdefault(server, {})
            settings.update(zip(names, IDs))
            # update the records for the packet
            for index, ID in zip(indices, IDs):
                records[index] = (ID,) + tuple(records[index][1:])

        returnValue((server, records))

    def clearCache(self):
        """Clear the cache of looked-up server and settings IDs."""
        self._serverCache = {}
        self._settingCache = {}

    def unflattenResponse(self, response_records):
        return [(ID, data.unflatten()) for (ID, data) in response_records]

    def _sendRequestNoLookup(self, target, records, context=(0, 0), timeout=None, unflatten=True):
        """Send a request without doing any lookups of server or setting IDs."""
        if self.disconnected:
            raise Exception('Already disconnected.')
        if len(self.pool):
            n = self.pool.pop()
        else:
            n = self._nextRequest
            self._nextRequest += 1

        self.requests[n] = d = defer.Deferred()
        if timeout is not None:
            timeoutCall = reactor.callLater(timeout, d.errback,
                                            errors.RequestTimeoutError())
            d.addBoth(self._cancelTimeout, timeoutCall)
        d.addBoth(self._finishRequest, n)
        if unflatten:
            d.addCallback(self.unflattenResponse)
        self.sendPacket(target, context, n, records)
        return d

    @inlineCallbacks
    def _sendManagerRequest(self, setting_id=None, data=None, timeout=None):
        if setting_id is None:
            records = []
        else:
            records = [(setting_id, data)]
        resp = yield self.sendRequest(C.MANAGER_ID, records, timeout=timeout)
        returnValue(resp[0][1])

    def _cancelTimeout(self, result, timeoutCall):
        """Cancel a pending request timeout call."""
        if timeoutCall.active():
            timeoutCall.cancel()
        return result

    def _finishRequest(self, result, n):
        """Finish a request."""
        del self.requests[n]
        self.pool.add(n) # reuse request numbers
        return result

    # receiving
    def dataReceived(self, data):
        self.packetStream.send(data)

    def packetReceived(self, source, context, request, records):
        """Process incoming packet."""
        if request > 0:
            self.requestReceived(source, context, request, records)
        elif request < 0:
            self.responseReceived(source, context, request, records)
        else:
            self.messageReceived(source, context, records)

    @inlineCallbacks
    def requestReceived(self, source, context, request, flat_records):
        """Process incoming request."""
        try:
            if self.request_handler is None:
                log.msg('server request_handler not set')
                raise Exception('server request_handler not set')
            response = yield self.request_handler(source, context, flat_records)
            self.sendPacket(source, context, -request, response)
        except Exception as e:
            # this will only happen if there was a problem while sending,
            # which usually means a problem flattening the response into
            # a valid LabRAD packet
            self.sendPacket(source, context, -request, [(0, e)])

    def responseReceived(self, source, context, request, flat_records):
        """Process incoming response."""
        if -request in self.requests: # reply has request number negated
            d = self.requests[-request]
            errors = [r[1].unflatten() for r in flat_records if isinstance(r[1].tag, T.LRError)]
            if errors:
                # fail on the first error
                d.errback(errors[0])
            else:
                d.callback(flat_records)
        else:
            # probably a response for a request that has already
            # timed out.  If not, something bad has happened.
            log.msg('Invalid response: %s, %s, %s, %s' % \
                    (source, context, request, records))

    def messageReceived(self, source, context, flat_records):
        """Process incoming messages."""
        self._messageLock.run(self._dispatchMessage, source, context, flat_records)

    @inlineCallbacks
    def _dispatchMessage(self, source, context, flat_records):
        """Dispatch a message to all matching listeners."""
        for ID, flat_data in flat_records:
            data = flat_data.unflatten()
            msgCtx = MessageContext(source, context, ID)
            keys = ((s, c, i) for s in (source, None)
                              for c in (context, None)
                              for i in (ID, None)
                              if (s, c, i) in self.listeners)
            for key in keys:
                for listener, async in self.listeners[key]:
                    func, args, kw = listener
                    try:
                        d = func(msgCtx, data, *args, **kw)
                        if not isinstance(d, defer.Deferred):
                            continue
                        if async:
                            d.addErrback(self._handleListenerError,
                                         msgCtx, data, listener)
                        else:
                            yield d
                    except Exception:
                        self._handleListenerError(
                            failure.Failure(), msgCtx, data, listener)

    def _handleListenerError(self, failure, msgCtx, data, listener):
        """Catch errors in message listeners.

        Just prints out a (hopefully informative) message
        about any errors that may have occurred.
        """
        print 'Unhandled error in message listener:', msgCtx, data, listener
        failure.printTraceback(elideFrameworkCode=1)

    # message handling
    def addListener(self, listener, source=None, context=None, ID=None, async=True, args=(), kw={}):
        """Add a listener for messages with the specified attributes.

        When a message with the specified source, context and ID is received,
        listener will be called with the message data, along with the args and
        keyword args specified here.  async determines how message listeners
        that return deferreds are to be handled.  If async is True, the message
        dispatcher will not wait for the deferred returned by a listener to fire.
        However, if async is False, the dispatcher will wait for this deferred
        before firing any more messages.
        """
        key = (source, context, ID)
        listeners = self.listeners.setdefault(key, [])
        listeners.append(((listener, args, kw), async))

    def removeListener(self, listener, source=None, context=None, ID=None):
        """Remove a listener for messages."""
        key = (source, context, ID)
        listeners = [l for l in self.listeners[key] if l[0][0] != listener]
        if len(listeners):
            self.listeners[key] = listeners
        else:
            del self.listeners[key]

    @inlineCallbacks
    def authenticate(self, username=None, password=None, headless=False):
        """Authenticate to the manager using the given credentials."""

        @inlineCallbacks
        def get_manager_auth_methods():
            if 'auth-server' in self.manager_features:
                # Ask the manager what auth methods it supports
                methods = yield self._sendManagerRequest(101, None)
            else:
                # Old managers only support password auth
                methods = ['password']
            returnValue(set(methods))

        def require_secure_connection(auth_type):
            is_secure = hasattr(self.transport, 'getPeerCertificate')
            is_local_connection = util.is_local_connection(self.transport)
            if not is_secure and not is_local_connection:
                raise Exception("cannot use {} auth over an insecure remote "
                                "connection".format(auth_type))

        # Get username and password from environment if not passed here
        if username is None:
            username = C.USERNAME
        if password is None:
            cred = auth.get_password(self.host, self.port, username,
                                     prompt=False)
            if cred is not None:
                password = cred.password

        if password is not None or username is not None:
            # Use password-based auth
            if username is None:
                # Have password; assume root user
                credential = auth.Password(password=password)
            elif password is None:
                # Have username; get password from cache or prompt
                credential = auth.get_password(
                        self.host, self.port, user=username, prompt=True)
            else:
                # Have username and password; use both
                if not username:
                    # Empty username is the root user
                    credential = auth.Password(password=password)
                else:
                    # Some other user; make sure manager supports user auth.
                    manager_auth_methods = yield get_manager_auth_methods()
                    if 'username+password' not in manager_auth_methods:
                        raise Exception('Manager does not support '
                                        'username+password auth. Cannot log in '
                                        'as user {}.'.format(username))
                    credential = auth.Password(username, password)

        else:
            # Have neither username nor password.
            # Check manager-supported auth methods; use OAuth if available.
            manager_auth_methods = yield get_manager_auth_methods()
            oauth_methods = {'oauth_token', 'oauth_access_token'}
            client_auth_methods = {'password', 'username+password'} | oauth_methods
            allowed = client_auth_methods & manager_auth_methods

            if oauth_methods & allowed:
                if 'oauth_access_token' in allowed:
                    # Prefer using access token if manager supports it.
                    method = 'oauth_access_token'
                else:
                    method = 'oauth_token'
                auth_info = yield self._sendManagerRequest(102, method)
                auth_info = dict(auth_info)
                credential = oauth.get_token(auth_info['client_id'],
                                             auth_info['client_secret'],
                                             headless=headless)
            elif 'username+password' in allowed:
                credential = auth.get_username_and_password(
                        self.host, self.port, prompt=True)
            else:
                credential = auth.get_password(
                        self.host, self.port, user='', prompt=True)

        if isinstance(credential, auth.Password):
            if not credential.username:
                # send login packet to get password challenge
                challenge = yield self._sendManagerRequest()

                # send password response
                m = hashlib.md5()
                m.update(challenge)
                m.update(credential.password)
                try:
                    resp = yield self._sendManagerRequest(0, m.digest())
                except Exception:
                    raise errors.LoginFailedError('Incorrect password.')
            else:
                method = 'username+password'
                require_secure_connection(method)
                try:
                    data = (credential.username, credential.password)
                    resp = yield self._sendManagerRequest(103, (method, data))
                except Exception as e:
                    raise errors.LoginFailedError(str(e))
            auth.cache_password(self.host, self.port, credential)

        elif isinstance(credential, oauth.OAuthToken):
            require_secure_connection(method)
            try:
                if method == 'oauth_access_token':
                    data = credential.access_token
                else:
                    data = credential.id_token
                resp = yield self._sendManagerRequest(103, (method, data))
            except Exception as e:
                raise errors.LoginFailedError(str(e))

        self.credential = credential
        self.loginMessage = resp

    def loginClient(self, name):
        """Log in as a client by sending our name to the manager.

        Args:
            name (str): The name of this labrad connection. Need not be unique.

        Returns:
            twisted.internet.defer.Deferred(None): A deferred that will fire
            after we have logged in.
        """
        return self._doLogin(name)

    def loginServer(self, name, descr, notes):
        """Log in as a server by sending our name and metadata to the manager.

        Args:
            name (str): The name of this server. Must be unique; login will
                fail if another server of the same name is already connected.
            descr (str): A description of this server, which will be exposed
                to other labrad clients.
            notes (str): More descriptive information about the server. This
                field is deprecated; instead we recommend just putting all info
                into descr.

        Returns:
            twisted.internet.defer.Deferred(None): A deferred that will fire
            after we have logged in.
        """
        return self._doLogin(name, descr, notes)

    @inlineCallbacks
    def _doLogin(self, *ident):
        # Store name, which is always the first identification param.
        self.name = ident[0]
        # Send identification.
        self.ID = yield self._sendManagerRequest(0, (1L,) + ident)


class MessageContext(object):
    """Object to be passed as the first argument to message handlers."""

    def __init__(self, source, context, target):
        self.source = source
        self.ID = context
        self.target = target

    def __repr__(self):
        return 'MessageContext(source=%s, ID=%s, target=%s)' % (self.source, self.ID, self.target)


# factory for creating LabRAD connections
_factory = protocol.ClientCreator(reactor, LabradProtocol)


@inlineCallbacks
def connect(host=C.MANAGER_HOST, port=None, tls_mode=C.MANAGER_TLS):
    """Connect to LabRAD and return a deferred that fires the protocol object.

    Args:
        host (str): The hostname of the manager.
        port (int): The tcp port of the manager. If None, use the appropriate
            default value based on the TLS mode.
        tls_mode (str): The tls mode to use for this connection. See:
            `labrad.constants.check_tls_mode`.

    Returns:
        twisted.internet.defer.Deferred(LabradProtocol): A deferred that will
        fire with the protocol once the connection is established.
    """
    tls_mode = C.check_tls_mode(tls_mode)
    if port is None:
        port = C.MANAGER_PORT_TLS if tls_mode == 'on' else C.MANAGER_PORT

    if tls_mode == 'on':
        tls_options = crypto.tls_options(host)
        p = yield _factory.connectSSL(host, port, tls_options, timeout=C.TIMEOUT)
        p.set_address(host, port)
        returnValue(p)

    @inlineCallbacks
    def do_connect():
        p = yield _factory.connectTCP(host, port, timeout=C.TIMEOUT)
        p.set_address(host, port)
        returnValue(p)

    @inlineCallbacks
    def start_tls(p, cert_string=None):
        try:
            cert = yield p._sendManagerRequest(1, ('STARTTLS', host))
        except Exception:
            raise Exception(
                'Failed sending STARTTLS command to server. You should update '
                'the manager and configure it to support encryption or else '
                'disable encryption for clients. See '
                'https://github.com/labrad/pylabrad/blob/master/CONFIG.md')
        p.transport.startTLS(crypto.tls_options(host, cert_string=cert_string))
        returnValue(cert)

    @inlineCallbacks
    def ping(p):
        resp = yield p._sendManagerRequest(2, 'PING')
        if isinstance(resp, tuple):
            manager_features = set(resp[1])
        else:
            manager_features = set()
        returnValue(manager_features)

    p = yield do_connect()
    is_local_connection = util.is_local_connection(p.transport)
    if ((tls_mode == 'starttls-force') or
        (tls_mode == 'starttls' and not is_local_connection)):
        try:
            cert = yield start_tls(p)
        except Exception:
            # TODO: remove this retry. This is a temporary fix to support
            # compatibility until TLS is fully deployed.
            print ('STARTTLS failed; will retry without encryption in case we '
                   'are connecting to a legacy manager.')
            p = yield connect(host, port, tls_mode='off')
            print 'Connected without encryption.'
            p.manager_features = set()
            returnValue(p)
        try:
            manager_features = yield ping(p)
        except Exception:
            print 'STARTTLS failed due to untrusted server certificate:'
            print 'SHA1 Fingerprint={}'.format(crypto.fingerprint(cert))
            print
            while True:
                ans = raw_input(
                        'Accept server certificate for host "{}"? (accept just '
                        'this [O]nce; [S]ave and always accept this cert; '
                        '[R]eject) '.format(host))
                ans = ans.lower()
                if ans in ['o', 's', 'r']:
                    break
                else:
                    print 'Invalid input:', ans
            if ans == 'r':
                raise
            p = yield do_connect()
            yield start_tls(p, cert)
            manager_features = yield ping(p)
            if ans == 's':
                # save now that we know TLS succeeded,
                # including hostname verification.
                crypto.save_cert(host, cert)
    else:
        manager_features = yield ping(p)
    p.manager_features = manager_features
    returnValue(p)



from __future__ import absolute_import

from concurrent.futures import Future
from twisted.internet import defer, reactor
from twisted.python import threadable
import pytest

import labrad.concurrent as concurrent
import labrad.thread as thread


def test_map_future():
    f = Future()
    mf = concurrent.map_future(f, lambda x: x + 1)
    f.set_result(1)
    assert mf.result() == 2


def test_map_future_on_already_completed_future():
    f = Future()
    f.set_result(1)
    mf = concurrent.map_future(f, lambda x: x + 1)
    assert mf.result() == 2


class TestCallFuture(object):
    """Test call_future function, which wraps twisted code in a Future."""

    @classmethod
    def setup_class(cls):
        # make sure twisted reactor is running
        thread.startReactor()

    def test_synchronous_func(self):
        def func():
            return threadable.isInIOThread()

        f = concurrent.call_future(func)
        assert f.result(timeout=0.1)

    def test_synchronous_error(self):
        def func():
            raise ValueError()

        f = concurrent.call_future(func)
        with pytest.raises(ValueError):
            f.result(timeout=0.1)

    def test_asynchronous_func(self):
        @defer.inlineCallbacks
        def func():
            d = defer.Deferred()
            reactor.callLater(0, d.callback, 'woot')
            result = yield d
            defer.returnValue(result)

        f = concurrent.call_future(func)
        assert f.result(timeout=0.1) == 'woot'

    def test_asynchronous_error(self):
        @defer.inlineCallbacks
        def func():
            d = defer.Deferred()
            reactor.callLater(0, d.callback, 'woot')
            result = yield d
            raise ValueError()

        f = concurrent.call_future(func)
        with pytest.raises(ValueError):
            f.result(timeout=0.1)


class TestFutureToDeferred(object):
    """Test future_to_deferred which wraps a Future in a twisted Deferred."""

    @classmethod
    def setup_class(cls):
        # make sure twisted reactor is running
        thread.startReactor()

    def test_result_already_set(self):
        @defer.inlineCallbacks
        def func(f):
            result = yield concurrent.future_to_deferred(f)
            defer.returnValue(result)

        f1 = Future()
        f1.set_result(1)
        f2 = concurrent.call_future(func, f1)
        assert f2.result(timeout=0.1) == 1

    def test_set_result_outside_reactor(self):
        @defer.inlineCallbacks
        def func(f):
            result = yield concurrent.future_to_deferred(f)
            defer.returnValue(result)

        f1 = Future()
        f2 = concurrent.call_future(func, f1)
        f1.set_result(1)
        assert f2.result(timeout=0.1) == 1

    def test_set_result_inside_reactor(self):
        @defer.inlineCallbacks
        def func(f):
            d = concurrent.future_to_deferred(f)
            f.set_result(1)
            result = yield d
            defer.returnValue(result)

        f = Future()
        future = concurrent.call_future(func, f)
        assert future.result() == 1

    def test_exception_already_set(self):
        @defer.inlineCallbacks
        def func(f):
            result = yield concurrent.future_to_deferred(f)
            defer.returnValue(result)

        f1 = Future()
        f1.set_exception(ValueError())
        f2 = concurrent.call_future(func, f1)
        with pytest.raises(ValueError):
            f2.result(timeout=0.1)

    def test_set_exception_outside_reactor(self):
        @defer.inlineCallbacks
        def func(f):
            result = yield concurrent.future_to_deferred(f)
            defer.returnValue(result)

        f1 = Future()
        f2 = concurrent.call_future(func, f1)
        f1.set_exception(ValueError())
        with pytest.raises(ValueError):
            f2.result(timeout=0.1)

    def test_set_result_inside_reactor(self):
        @defer.inlineCallbacks
        def func(f):
            d = concurrent.future_to_deferred(f)
            f.set_exception(ValueError())
            result = yield d
            defer.returnValue(result)

        f = Future()
        future = concurrent.call_future(func, f)
        with pytest.raises(ValueError):
            future.result()


if __name__ == '__main__':
    pytest.main(['-v', __file__])



import math

import pytest

import labrad.types as T
from labrad.units import Complex, Value


def test_eval_value_with_none_for_units():
    assert T.evalLRData("""Value(1.0, None)""") == Value(1.0, '')


def test_eval_complex_with_none_for_units():
    assert T.evalLRData("""Complex(1.1, None)""") == Complex(1.1, '')


def test_eval_cluster_with_value_and_complex():
    assert (T.evalLRData("""(Value(1.0, None), Complex(2.0, None))""") ==
            (Value(1.0, ''), Complex(2.0, '')))


def test_eval_nan():
    s = T.reprLRData(float('nan'))
    n = T.evalLRData(s)
    assert math.isnan(n)

    s2 = T.reprLRData((float('nan'), 'this is a test'))
    n2, _ = T.evalLRData(s2)
    assert math.isnan(n2)


def test_eval_positive_infinity():
    s = T.reprLRData(float('inf'))
    n = T.evalLRData(s)
    assert math.isinf(n)
    assert n > 0

    s2 = T.reprLRData((float('inf'), 'this is a test'))
    n2, _ = T.evalLRData(s2)
    assert math.isinf(n2)
    assert n2 > 0


def test_eval_negative_infinity():
    s = T.reprLRData(float('-inf'))
    n = T.evalLRData(s)
    assert math.isinf(n)
    assert n < 0

    s2 = T.reprLRData((float('-inf'), 'this is a test'))
    n2, _ = T.evalLRData(s2)
    assert math.isinf(n2)
    assert n2 < 0


if __name__ == '__main__':
    pytest.main(['-v', __file__])



# Copyright (C) 2007  Matthew Neeley
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
labrad.server

Base classes for building asynchronous, context- and request- aware
servers with labrad.
"""

from __future__ import absolute_import

from datetime import datetime
from operator import attrgetter
import threading
import traceback

from concurrent import futures
from twisted.internet import defer, reactor, threads
from twisted.internet.defer import inlineCallbacks, returnValue
from twisted.internet.error import ConnectionDone, ConnectionLost
from twisted.python import failure, log, threadable

from labrad import constants as C, types as T, util
import labrad.backend
import labrad.client
import labrad.concurrent
from labrad.decorators import setting
from labrad.wrappers import ClientAsync


class Signal(object):
    """A Signal object is a simple publish/subscribe messaging primitive.

    Servers expose signals as settings that clients or other servers can
    make requests to.  A request allows the client or server to sign up to
    receive messages when this signal is fired.  All the details of managing
    who is listening for these signals is handled here, so the application
    does not need to worry about it.  The application can simply call the
    signal object and the data given will be sent to all connected listeners.
    """
    def __init__(self, ID, name, returns=[]):
        self.ID = ID
        self.name = name
        self.tag = returns
        self.listeners = {}

    def __call__(self, data, contexts=None, tag=None):
        """Fire a message with the given data to all connected listeners.

        If no context is specified, the message will be sent to all listeners.
        Otherwise, the message will be sent only to those listeners signed
        up in the context or contexts specified.
        """
        cxn = self.parent._cxn
        if contexts is None:
            # send this to everyone
            for context, targets in self.listeners.items():
                for target, ID in targets.items():
                    cxn.sendMessage(target, [(ID, data, self.tag)], context)
        else:
            # this little hack allows one to specify contexts
            # as either a single context tuple, or an iterable of
            # multiple context tuples.
            if isinstance(contexts, tuple) and len(contexts) \
               and not isinstance(contexts[0], tuple):
                contexts = [contexts]
            # send only to those in the specified
            # context or list of contexts
            for context in contexts:
                if context not in self.listeners:
                    continue
                for target, ID in self.listeners[context].items():
                    cxn.sendMessage(target, [(ID, data, self.tag)], context)

    def connect(self, context, target, ID):
        """Connect a listener to this signal.

        Connections are made to a given target/message ID, in a particular
        context.  When the signal is fired later, it can be fired to all listeners,
        or only to those listeners signed up in particular contexts.
        """
        cdict = self.listeners.setdefault(context, {})
        tdict = cdict.setdefault(target, ID)

    def disconnect(self, context=None, target=None):
        """Disconnect listeners identified by context and target."""
        # disconnect a particular target in a particular context
        if context is not None and target is not None:
            if context in self.listeners:
                targets = self.listeners[context]
                if target in targets:
                    del targets[target]
                if not len(targets):
                    del self.listeners[context]
        # disconnect a particular target in any context
        elif target is not None:
            for context, targets in self.listeners.items():
                if target in targets:
                    del targets[target]
                if not len(targets) or context[0] == target:
                    del self.listeners[context]
        # disconnect a particular context
        elif context is not None:
            if context in self.listeners:
                del self.listeners[context]
        # disconnect all listeners
        else:
            self.listeners = {}

    # implement the IRequestHandler interface for connect/disconnect
    def handleRequest(self, server, c, data):
        return self._handler.handleRequest(server, c, data)

    def getRegistrationInfo(self):
        return self._handler.getRegistrationInfo()

    @property
    def returns(self):
        return self._handler.returns

    @property
    def _handler(self):
        """Request handler to manage connect/disconnect requests."""
        if hasattr(self, '_handler_func'):
            return self._handler_func
        @setting(self.ID, self.name, ID=['w'], returns=[''])
        def handler(server, c, ID=None):
            if ID is not None:
                self.connect(c.ID, c.source, ID)
            else:
                self.disconnect(c.ID, c.source)
        handler.__doc__ = """Connect to/Disconnect from signal '%s'

        Passing a word ID will cause messages to be sent to setting ID
        when this signal is fired.  Passing nothing will cancel future messages.
        The message data will be of type '%s'.""" % (self.name, self.tag)
        self._handler_func = handler
        return self._handler_func


class Context(object):
    """Serialize requests in a context and handle context expiration.

    One Context object is created for each request context seen by a
    server.  This has a lock so that requests in the context will happen
    in series, and it also stores context data that will be passed
    in to handlers.  In addition, this object keeps track of whether
    the context has expired, so that requests in expired contexts will
    be terminated properly.
    """

    def __init__(self):
        self.lock = defer.DeferredLock()
        self.expired = False

    def acquire(self):
        """Acquire the lock on this context."""
        self.check()
        return self.lock.acquire()

    def check(self):
        """Check to see whether this context has expired."""
        if self.expired:
            raise Exception("Context expired.")

    def release(self):
        """Release the lock on this context."""
        self.lock.release()

    def expire(self):
        """Mark this context as having expired."""
        self.expired = True


class LabradServer(object):
    """A generic LabRAD server."""

    sendTracebacks = True
    prioritizeWrites = False

    def __init__(self):
        self.description, self.notes = util.parseSettingDoc(self.__doc__)

        self.started = False
        self.stopping = False

        self.onStartup = util.DeferredSignal()
        self.onShutdown = util.DeferredSignal()

        self.settings = {}
        self.signals = []
        self.contexts = {}

        self.__protocol = None
        self.__async_client = None
        self.__thread_data = threading.local()

    def _dispatch(self, func, *args, **kw):
        return func(*args, **kw)

    # request handling
    @inlineCallbacks
    def request_handler(self, source, context, flat_records):
        """Handle an incoming request.

        If this is a new context, we create a context object and a lock
        to ensure that requests in this context are serialized.  Then,
        we serve the request.
        """
        # create a new context if needed
        c = self.contexts.get(context, None)
        if c is None:
            c = self.contexts[context] = Context()
            yield c.acquire() # make sure we're the first in line
            c.data = yield self._dispatch(self.newContext, context)
            yield self._dispatch(self.initContext, c.data)
        else:
            yield c.acquire() # wait for previous requests in this context to finish

        if self.prioritizeWrites:
            # yield here so that pending writes will be sent in preference to processing
            # new requests.  This can help in cases where server settings do long-running
            # computations that block, though we are still limited fundamentally by
            # the completely single-threaded way twisted operates
            yield util.wakeupCall(0.0)
        response = []
        try:
            yield self._dispatch(self.startRequest, c.data, source)
            for ID, flat_data in flat_records:
                c.check() # make sure this context hasn't expired
                try:
                    setting = self.settings[ID]
                    result = yield self._dispatch(setting.handleRequest, self, c.data, flat_data)
                    response.append((ID, result, setting.returns))
                except Exception, e:
                    response.append((ID, self._getTraceback(e)))
                    break
            c.check() # make sure this context hasn't expired
        finally:
            reactor.callLater(0, c.release)
        returnValue(response)

    def _getTraceback(self, e):
        """Turn an exception into a LabRAD error packet.

        We may send a traceback of the error, depending
        on the value of self.sendTracebacks.
        """
        code = getattr(e, 'code', 0)
        if self.sendTracebacks:
            f = failure.Failure()
            tb = f.getTraceback(elideFrameworkCode=True)
            msg = 'Remote %s' % tb
        else:
            msg = e.__class__.__name__ + ': ' + getattr(e, 'msg', str(e))
        msg = '[%s] %s' % (self.name, msg)
        return T.Error(msg, code)

    # registering setting and signal handlers
    def _findSettingHandlers(self):
        """Find all settings defined for this server."""
        # this is an ad-hoc test; we really should check for the IRequestHandler interface
        members = [getattr(self, name) for name in dir(self)]
        handlers = [m for m in members if hasattr(m, 'handleRequest')]
        return sorted(handlers, key=attrgetter('ID'))

    def _checkSettingConflicts(self, s):
        """Check for conflicts in setting name and ID."""
        for other in self.settings.values():
            if s.ID == other.ID or s.name == other.name:
                msg = "Conflicting settings: '%s' (%d), '%s' (%d)" % \
                      (s.name, s.ID, other.name, other.ID)
                raise AssertionError(msg)

    def addSetting(self, setting, packet=None):
        """Add a new setting to the server.

        If a packet is given, the registration call will be added to the
        packet, otherwise a request will be made to the manager, and we
        return a deferred that will fire when the request is done.
        """
        self._checkSettingConflicts(setting)
        self.settings[setting.ID] = setting
        if packet is None:
            packet = self.client.manager
        info = setting.getRegistrationInfo()
        return packet.s__register_setting(info)

    def removeSetting(self, setting, packet=None):
        """Remove a setting from the server.

        If a packet is given, the deregistration call will be added to the
        packet, otherwise a request will be made to the manager, and we
        return a deferred that will fire when the request is done.
        """
        del self.settings[setting.ID]
        if packet is None:
            packet = self.client.manager
        return packet.s__unregister_setting(setting.ID)

    def addSignal(self, signal, packet=None):
        """Add a new signal to the server.

        We add it to the list of signals and also add the
        request handler for signing up to this signal.
        """
        signal.parent = self
        self.signals.append(signal)
        return self.addSetting(signal, packet)

    def removeSignal(self, signal, packet=None):
        """Remove a signal from this server.

        We remove it from the list of signals and also remove
        the request handler for signing up for this signal.
        """
        self.signals.remove(signal)
        return self.removeSetting(signal, packet)

    @inlineCallbacks
    def startup(self, protocol):
        """Start this server using the given protocol connection.

        Identifies this server to the manager, creates an async wrapper for the
        protocol connection, and then runs initialization callbacks for this
        server.

        Args:
            protocol (labrad.protocol.LabradProtocol): A protocol connection
                to the labrad manager, as returned by labrad.protocol.connect.
                This protocol must have been authenticated prior to calling
                startup.

        Returns:
            twisted.internet.defer.Deferred(None): A deferred that will fire
            once startup is complete.
        """
        try:
            name = getattr(self, 'instanceName', self.name)
            yield protocol.loginServer(name, self.description, self.notes)
            self._cxn = protocol
            self.ID = protocol.ID
            protocol.request_handler = self.request_handler
            protocol.onDisconnect().addBoth(self._connection_lost)
            self.__protocol = protocol
            self.__async_client = ClientAsync(protocol)
            yield self.__async_client._init()
            yield self._initServer()
            self.started = True
            self.onStartup.callback(self)
        except Exception as e:
            log.err("connection failed, disconnecting")
            traceback.print_exc()
            self.disconnect(e)
            raise

    @property
    def client(self):
        """Get a labrad client for this server's labrad connection.

        To accomodate asynchronous and synchronous (threaded) server
        implementations, this returns a single shared AsyncClient instance if
        called from the twisted reactor thread, or a thread-local Client
        instance for synchronous use if called from any other thread.
        """
        if threadable.isInIOThread():
            # We're in the reactor thread, so can return shared async client.
            return self.__async_client
        else:
            data = self.__thread_data
            if not hasattr(data, 'sync_client'):
                backend = labrad.backend.TwistedConnection(self.__protocol)
                data.sync_client = labrad.client.Client(backend)
            return data.sync_client

    # Network events
    # these methods are called by network events from twisted

    @inlineCallbacks
    def _initServer(self):
        """Called after we've authenticated with the LabRAD manager.

        Here we register the settings and signals found on this server
        and set up message handlers for messages coming from the manager.
        """
        log.msg('%s starting...' % self.name)
        # register handlers for settings and signals
        mgr = self.client.manager
        p = mgr.packet()
        for s in self._findSettingHandlers():
            if isinstance(s, Signal):
                self.addSignal(s, p)
            else:
                self.addSetting(s, p)

        yield p.send()

        # do server-specific initialization
        yield self._dispatch(self.initServer)
        # make sure we shut down gracefully when reactor quits or a remote message is fired
        self._shutdownID = reactor.addSystemEventTrigger('before', 'shutdown', self._stopServer)
        self._cxn.addListener(self._stopServer, ID=987654321)

        # sign up for notifications from the manager
        yield mgr.subscribe_to_named_message('Server Connect', 55443322, True)
        yield mgr.subscribe_to_named_message('Server Disconnect', 66554433, True)
        self._cxn.addListener(self._serverConnected, source=mgr.ID, ID=55443322, async=False)
        self._cxn.addListener(self._serverDisconnected, source=mgr.ID, ID=66554433, async=False)

        #yield mgr.notify_on_connect.connect(self._serverConnected)
        #yield mgr.notify_on_disconnect.connect(self._serverDisconnected)
        yield mgr.s__notify_on_context_expiration.connect(
                  self._contextExpired, connectargs=(True,))

        # let the rest of the world know we're ready
        yield mgr.s__start_serving()
        log.msg('%s now serving' % self.name)

    @inlineCallbacks
    def _stopServer(self, *ignored):
        self.stopping = True
        try:
            yield self._dispatch(self.stopServer)
        except Exception, e:
            self._error = failure.Failure(e)
        finally:
            try:
                self._cxn.disconnect()
            except Exception:
                pass
            # remove the event trigger, so we don't get
            # called again if the reactor shuts down later
            if hasattr(self, '_shutdownID'):
                reactor.removeSystemEventTrigger(self._shutdownID)

    def _connection_lost(self, reason):
        """Called when the network connection to LabRAD is lost.

        This could happen either because there was an error and the
        connection was lost prematurely, or because we requested a shutdown.
        """
        reason = getattr(self, '_error', reason)
        if not self.started:
            self.onStartup.errback(reason)
        else:
            if self.stopping and (reason is None or reason.check(ConnectionDone, ConnectionLost)):
                self.onShutdown.callback()
            else:
                self.onShutdown.errback(reason)

    def disconnect(self, error=None):
        """Disconnect this server from LabRAD.

        This method returns nothing, but you can call onShutdown()
        to get a deferred that will fire when the shutdown is done.
        """
        if error is not None:
            if not isinstance(error, failure.Failure):
                error = failure.Failure(error)
            self._error = error
        self._stopServer()

    # Server startup and shutdown
    # these methods should be overridden
    def initServer(self):
        """Initialize Server.

        Called after registering settings and creating a client
        connection to labrad, but before we start serving requests.
        """

    def stopServer(self):
        """Stop the server.

        Called when the server is shutting down, but before we have
        closed any client connections.  Perform any cleanup operations here.
        """

    # Connect/Disconnect notifications
    # these methods should not be overridden
    def _serverConnected(self, c, data):
        """Handle connection messages from the manager."""
        ID, name = data
        self._dispatch(self.serverConnected, ID, name)

    def _serverDisconnected(self, c, data):
        """Handle disconnection messages from the manager."""
        ID, name = data
        self._dispatch(self.serverDisconnected, ID, name)

    # these methods should be overridden
    def serverConnected(self, ID, name):
        """Called when a new server connects to LabRAD."""

    def serverDisconnected(self, ID, name):
        """Called when a server disconnects from LabRAD."""


    # Context handling
    # these methods should not be overridden
    def _contextExpired(self, c, data):
        """Handle context expiration messages.

        This gets called by the manager either with a single word,
        to expire all contexts with that high word, or with a tuple
        of high word, low word, to expire just that context.
        """
        if isinstance(data, tuple):
            # context tuple specified
            # expire only this context
            for signal in self.signals:
                signal.disconnect(context=data)
            self._expireContext(data)
        else:
            # client ID specified
            # expire all contexts for this ID
            for signal in self.signals:
                signal.disconnect(target=data)
            self._expireID(data)

    def _expireID(self, ID):
        """Expire all contexts with a given ID (high word)."""
        for context in self.contexts.keys():
            if context[0] == ID:
                self._expireContext(context)

    def _expireContext(self, context):
        """Expire a single context."""
        if context in self.contexts:
            c = self.contexts[context]
            del self.contexts[context]
            c.expire()
            self._dispatch(self.expireContext, c.data)

    # these methods may be overridden
    def newContext(self, ID):
        """Create a new context object."""
        c = util.ContextDict()
        c.ID = ID
        return c

    def initContext(self, c):
        """Initialize a new context object."""

    def startRequest(self, c, source):
        """Start a request from source in the given context."""
        c.source = source

    def expireContext(self, c):
        """Expire Context.

        Called when the client that created a context disconnects.
        Any cleanup operations on the context should be done here.
        """


    # Handlers
    # these handle remotely-accessible settings
    @setting(11111111)
    def echo(self, c, data):
        """Echo any packet."""
        return data

    @setting(12121212, returns=['s'])
    def debug(self, c, data):
        """Get a __repr__ of the current context."""
        return repr(c)


    # Signals
    onLog = Signal(13131313, 'signal: log', 't*s')

    def log(self, *messages):
        self.onLog((datetime.now(), messages))


class ThreadedServer(LabradServer):
    """A LabRAD server that dispatches requests to a thread pool."""

    def __init__(self, pool=None):
        """Create a new threaded server.

        Requests and lifecycle methods like initServer will be executed on a
        thread pool instead of in the twisted reactor thread. In addition,
        accessing self.client from a thread other than the reactor thread will
        return a synchronous labrad.client.Client object.

        Args:
            pool (None | concurrent.futures.ThreadPoolExecutor):
                Thread pool instance to use for server lifecycle methods and
                request handling. If None, use the default twisted threadpool,
                which maxes out at 10 threads.
        """
        super(ThreadedServer, self).__init__()
        self.__pool = pool

    @inlineCallbacks
    def _dispatch(self, func, *args, **kw):
        if self.__pool is None:
            result = yield threads.deferToThread(func, *args, **kw)
        else:
            result = self.__pool.submit(func, *args, **kw)
        while isinstance(result, futures.Future):
            result = yield labrad.concurrent.future_to_deferred(result)
        returnValue(result)


class SingleThreadedServer(ThreadedServer):
    """A LabRAD server that handles requests in a single thread."""
    def __init__(self):
        pool = futures.ThreadPoolExecutor(max_workers=1)
        super(SingleThreadedServer, self).__init__(pool)




# Copyright (C) 2007  Matthew Neeley
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


# Physical quantities with units
#
# Written by Konrad Hinsen <hinsen@cnrs-orleans.fr>
# with contributions from Greg Ward
# last revision: 2007-5-25


"""
Physical quantities with units.

This module provides a data type that represents a physical
quantity together with its unit. It is possible to add and
subtract these quantities if the units are compatible, and
a quantity can be converted to another compatible unit.
Multiplication, subtraction, and raising to rational powers
is allowed without restriction, and the result will have
the correct unit.

The values of physical constants are taken from the 1986
recommended values from CODATA. Other conversion factors
(e.g. for British units) come from various sources. I can't
guarantee for the correctness of all entries in the unit
table, so use this at your own risk.

This module is based on code from the ScientificPython package.
The version included with LabRAD has been slightly changed:

    - removed depency on numpy or numeric
    - included NumberDict so the module is self-contained
    - units that cannot be converted to SI are kept as text alone
    - physical quantity types are derived from numeric types, and the
      WithUnit mixin class.  Hence, they can be used directly as numbers
    - Value and Complex types define a __getitem__ method to
      express the value in a new unit, e.g:
      >>> Value(1, 'GHz')['Hz'] -> 1e9

"""

from fractions import Fraction
from math import floor, pi

import numpy as np

from labrad import grammar
from labrad.util import cache

# Dictionary containing numbers
#
# These objects are meant to be used like arrays with generalized
# indices. Non-existent elements default to zero. Global operations
# are addition, subtraction, and multiplication/division by a scalar.
#
# Written by Konrad Hinsen <hinsen@cnrs-orleans.fr>
# last revision: 2006-10-16

class NumberDict(dict):
    """
    Dictionary storing numerical values

    Constructor: NumberDict()

    An instance of this class acts like an array of number with
    generalized (non-integer) indices. A value of zero is assumed
    for undefined entries. NumberDict instances support addition,
    and subtraction with other NumberDict instances, and multiplication
    and division by scalars.
    """

    def __getitem__(self, item):
        try:
            return dict.__getitem__(self, item)
        except KeyError:
            return 0

    def __coerce__(self, other):
        if isinstance(other, dict):
            other = NumberDict(other)
        return self, other

    def __neg__(self):
        for key in self:
            self[key] *= -1

    def __add__(self, other):
        sum_dict = NumberDict()
        for key in self.keys():
            sum_dict[key] = self[key]
        for key in other.keys():
            sum_dict[key] = sum_dict[key] + other[key]
        return sum_dict

    def __sub__(self, other):
        sum_dict = NumberDict()
        for key in self.keys():
            sum_dict[key] = self[key]
        for key in other.keys():
            sum_dict[key] = sum_dict[key] - other[key]
        return sum_dict

    def __mul__(self, other):
        new = NumberDict()
        for key in self.keys():
            new[key] = other * self[key]
        return new

    __rmul__ = __mul__

    def __div__(self, other):
        new = NumberDict()
        for key in self.keys():
            new[key] = self[key] / other
        return new


class Lazy(object):
    """A method decorator which converts methods to lazy properties.

    A lazy property is computed on demand the first time it is accessed,
    and the result of the computation is then stored in the object's instance
    __dict__ so that subsequent access is just a dict lookup and incurs no
    overhead from descriptor access.

    Example use:

    class Foo(object):
        @Lazy
        def bar(self):
            return intense_computation(self, ...)

    >>> foo = Foo()
    >>> result_first_access = foo.bar
    # Incurs long computation. Note the lack of ().

    >>> result_second_access = foo.bar
    # Incurs no computation. Looks up previously computed and cached result.

    Methods decorated by Lazy must take no arguments aside from the implicit
    self argument and must return a deterministic result. Also note that the
    decorated method is accessed as an attribute, i.e. no ().

    Attributes:
        f (function): The decorated function. f is called only the first time
            the decorated method is accessed.
    """

    def __init__(self, f):
        """Initialize a Lazy property

        Args:
            f (function): The function to by lazified.

        Returns nothing. However, Lazy is used as a decorator on methods, so
        Lazy.__new__ is called automatically at class creation time and returns
        a Lazy instance which wraps the original method f.
        """
        self.f = f

    def __get__(self, obj, objtype=None):
        result = self.f(obj)
        obj.__dict__[self.f.__name__] = result
        return result


class WithUnit(object):
    """Mixin class for adding units to numeric types.

    Quantities with units can be used with basic arithmetic operations
    (addition, subtraction, multiplication, division, raising to powers).

    The 'unit' property stores the units for a given quantity, and you
    can get the bare value converted to a particular unit 'u' by using
    square-bracket indexing with the name of a unit or a unit object:

    >>> x = Value(1, 's')
    >>> x
    Value(1.0, 's')
    >>> x.unit
    <Unit 's'>
    >>> x['ms']
    1000.0

    To just strip off the units, you can use the square bracket indexing
    with the 'unit' property. For example, if you wanted to print the value
    and units separately:

    >>> print "The value is {:3.2f} {}".format(x[x.unit], x.unit)
    1.00 s
    """

    __array_priority__ = 15

    def __new__(cls, value, unit=None):
        # Convert inputs to make sense:
        # check for illegal None unit, unit string -> Unit, list -> array, check
        # to see if value is
        if unit is None:
            raise RuntimeError("Cannot construct WithUnit with unit=None.  Use correct units, or use a float")
        if isinstance(value, list):
            value = np.asarray(value)
        unit = Unit(unit)
        if hasattr(value, 'unit'): # This is called from Unit * Value and friends
            return WithUnit(value._value, value.unit * unit)

        # Ok, with all that business taken care of, look up the right type
        # (Value, Complex, ValueArray, or the Dimensionless versions of each
        # and construct a final object
        cls = cls._findClass(type(value), unit)
        if unit and unit.is_dimensionless:
            return cls(value * unit.conversionFactorTo(''))
        inst = super(WithUnit, cls).__new__(cls)
        inst._value = inst._numType(value) * 1.0 # For numpy: int to float
        inst.unit = Unit(unit)
        inst.is_dimensionless = inst.unit is None or inst.unit.is_dimensionless
        return inst

    _numericTypes = {}
    _dimensionlessTypes = {}

    @classmethod
    def _findClass(cls, numType, unit):
        """Find a class for a particular numeric type and unit.

        Classes will be created if they haven't been already.
        """
        # find class with units for this numeric type
        try:
            if unit.is_dimensionless:
                cls = cls._dimensionlessTypes[numType]
            else:
                cls = cls._numericTypes[numType]
        except KeyError:
            raise TypeError('Cannot use units with instances of type %s' % (numType,))
        return cls

    def __reduce__(self):
        return (WithUnit, (self._value, self.unit.name))

    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        return self

    def __str__(self):
        return '%s %s' % (self._value, self.unit)

    def __repr__(self):
        return '%s(%r, %r)' % (self.__class__.__name__, self._value, self.unit.name)

    @property
    def units(self):
        """A string representation of the unit of this value."""
        return self.unit.name

    def _convert_units(self, other):
        """Convert two objects to use compatible units.

        This is used by __add__ and __sub__ as well as the __lt__ family of comparison
        operators.  This repalces the old __cmp__ based operation which didn't work on
        inf and nans.

        This returns (left, right, unit) where 'left' and 'right' are bare numeric types
        (float, complex, array), and unit is their common unit.
        """
        if isinstance(other, WithUnit):
            if self.isCompatible(other):
                if self.unit.conversionFactorTo(other.unit) > 1:
                    unit = other.unit
                else:
                    unit = self.unit
                return (self[unit], other[unit], unit)
            else:
                raise TypeError("Incompatible units: %s, %s" % (self.unit, other.unit))
        elif self.is_dimensionless:
            return (self[''], other, U.Unit(''))
        elif other == 0:
            return (self._value, 0.0, self.unit)
        else:
            raise TypeError("Incompatible Units %s, ''" % (self.unit,))

    def __add__(self, other):
        a, b, unit = self._convert_units(other)
        return WithUnit(a+b, unit)

    __radd__ = __add__

    def __sub__(self, other):
        a, b, unit = self._convert_units(other)
        return WithUnit(a-b, unit)

    def __rsub__(self, other):
        a, b, unit = self._convert_units(other)
        return WithUnit(b-a, unit)

    def __eq__(self, other):
        try:
            a, b, unit = self._convert_units(other)
            return a==b
        except Exception as e:
            # Anything that prevents comparison means and b are not equal
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        a, b, unit = self._convert_units(other)
        return a < b

    def __le__(self, other):
        a, b, unit = self._convert_units(other)
        return a <= b

    def __gt__(self, other):
        a, b, unit = self._convert_units(other)
        return a > b

    def __ge__(self, other):
        a, b, unit = self._convert_units(other)
        return a >= b

    def __mul__(self, other):
        if isinstance(other, Unit):
            return WithUnit(self._value, self.unit * other)
        if isinstance(other, WithUnit):
            return WithUnit(self._value * other._value,
                            self.unit * other.unit)
        return WithUnit(self._value * other, self.unit)

    __rmul__ = __mul__

    def __div__(self, other):
        if isinstance(other, Unit):
            return WithUnit(self._value, self.unit / other)
        if isinstance(other, WithUnit):
            return WithUnit(self._value / other._value,
                            self.unit / other.unit)
        return WithUnit(self._value / other, self.unit)

    __truediv__ = __div__

    def __rdiv__(self, other):
        if isinstance(other, Unit):
            return WithUnit(self._value, other / self.unit)
        if isinstance(other, WithUnit):
            return WithUnit(other._value / self._value,
                            other.unit / self.unit)
        return WithUnit(other / self._value, pow(self.unit, -1))

    __rtruediv__ = __rdiv__

    def __pow__(self, other):
        if isinstance(other, (WithUnit, Unit)) and not other.is_dimensionless:
            raise TypeError('Exponents must be dimensionless')
        return WithUnit(pow(self._value, float(other)),
                        pow(self.unit, other))

    def __rpow__(self, other):
        if not self.is_dimensionless:
            raise TypeError('Exponents must be dimensionless')
        return pow(other, self._value * self.unit.factor)

    def __abs__(self):
        return WithUnit(abs(self._value), self.unit)

    def __pos__(self):
        return self

    def __neg__(self):
        return WithUnit(-self._value, self.unit)

    def __nonzero__(self):
        return self._value != 0

    def __hash__(self):
        '''
        Not sure if using Value as a dictionary key makes sense, but
        here it is anyway.  We convert to base units so that 1*km and
        1000*m hash to the same value.  They also hash to the same
        value as 1000*V, but that is OK.  Note that this is still kind
        of messed up: 1001*m != 1.001*km because the latter is not an
        exact number.  C'est la floating point.
        '''
        if not hasattr(self, '_hash'):
            self._hash = hash(self._base_value)
        return self._hash

    @Lazy
    def _base_value(self):
        """The bare value of this WithUnit in units of our base unit.

        This is used internally for performance reasons in certain contexts,
        but should not be used externally.
        """
        return self[self.unit.base_unit]

    def __getitem__(self, unit):
        """Return value of physical quantity expressed in new units."""
        if unit == self.unit:
            return self._value
        factor, offset = self.unit.conversionTupleTo(unit)
        return (self._value + offset) * factor

    def inUnitsOf(self, unit):
        """
        Express the quantity in different units. If one unit is
        specified, a new PhysicalQuantity object is returned that
        expresses the quantity in that unit. If several units
        are specified, the return value is a tuple of
        PhysicalObject instances with with one element per unit such
        that the sum of all quantities in the tuple equals the the
        original quantity and all the values except for the last one
        are integers. This is used to convert to irregular unit
        systems like hour/minute/second.

        @param unit: one or several units
        @type unit: C{str}
        @returns: one or more physical quantities
        @rtype: L{PhysicalQuantity} or C{tuple} of L{PhysicalQuantity}
        @raises TypeError: if any of the specified units are not compatible
        with the original unit
        """
        if unit == self.unit:
            return self
        u = Unit(unit)
        return self[u] * u

    # Contributed by Berthold Hoellmann
    def inBaseUnits(self):
        """
        @returns: the same quantity converted to base units,
        i.e. SI units in most cases
        @rtype: L{PhysicalQuantity}
        """
        return self.inUnitsOf(self.unit.base_unit)

    def isCompatible(self, unit):
        """
        @param unit: a unit
        @type unit: C{str}
        @returns: C{True} if the specified unit is compatible with the
        one of the quantity
        @rtype: C{bool}
        """
        return self.unit.isCompatible(unit)

    def isDimensionless(self):
        return self.is_dimensionless

    def sqrt(self):
        return np.sqrt(self._value) * pow(self.unit, Fraction(1, 2))


class Value(WithUnit):
    _numType = float

    def __float__(self):
        return self['']

    def __iter__(self):
        raise TypeError("'Value' object is not iterable")

WithUnit._numericTypes[float] = Value
WithUnit._numericTypes[int] = Value
WithUnit._numericTypes[long] = Value
WithUnit._numericTypes[np.int32] = Value
WithUnit._numericTypes[np.int64] = Value


class Complex(WithUnit):
    _numType = complex

    def __complex__(self):
        return self['']

    def __iter__(self):
        raise TypeError("'Complex' object is not iterable")

WithUnit._numericTypes[complex] = Complex


class ValueArray(WithUnit):

    _numType = np.array # Regular ndarray constructor doesn't work

    def __new__(cls, data, unit=None):
        '''
        unit=None can only be used if data is an iterable of items that already have units
        '''
        if unit is not None:
            return super(ValueArray, cls).__new__(cls, data, unit)

        it = iter(data)
        first = it.next()
        unit = first.unit
        first = first[unit] # convert to float
        rest = [x[unit] for x in it]
        return WithUnit([first] + rest, unit)

    def __copy__(self):
        # Numpy arrays are not immutable so we have to
        # make a real copy
        return WithUnit(self._value.copy(), self.unit)

    def __deepcopy__(self, memo):
        return self.__copy__()

    def __getitem__(self, unit):
        if isinstance(unit, (str, Unit)):
            """Return value of physical quantity expressed in new units."""
            return super(ValueArray, self).__getitem__(unit)
        else:
            idx = unit
            return WithUnit(self._value[idx], self.unit)

    def __setitem__(self, key, value):
        self._value[key] = value.inUnitsOf(self.unit)._value

    def __len__(self):
        return len(self._value)

    # expose useful attributes from the backing array

    @property
    def dtype(self):
        return self._value.dtype

    @property
    def ndim(self):
        return self._value.ndim

    @property
    def shape(self):
        return self._value.shape

    def allclose(self, other, *args, **kw):
        return np.allclose(self._value, other[self.unit], *args, **kw)

WithUnit._numericTypes[np.ndarray] = ValueArray
WithUnit._numericTypes[list] = ValueArray

# add support for numeric types returned by most numpy/scipy functions
WithUnit._numericTypes[np.float64] = Value
WithUnit._numericTypes[np.complex128] = Complex

# Unit class implementation
#
# Unit class attributes:
#       powers:         9 element list of the power of each SI base unit
#       lex_names:      Number Dictionary of non-SI "base" units (i.e., {'counts': 1}).  
#       names:          (Number)-Dictionary of display names with the associated powers
#       factor:         Ratio between the unit value described by 'names', and the base
#                       units described by powers and lex_names.
#       offset:         Offset from zero used for degF and degC.
#
# So, for example, the unit representation of a few types are given here:
#
# N [newton]: names: {'N': 1}, powers: [1, 1, -2, 0, ...] lex_names: {}, factor: 1.0
# l/s [liter/sec]: names: {'l': 1, 's': -1} powers: [3, 0, -1, 0...] lex_names: {}, factor: .001
# kN: names: {'kN': 1}, powers: [1, 1, -2, 0, ...], lex_names: {}, factor: 1000
# 'count': names: {'count': 1}, powers: [0, 0, 0,...] lex_names: {'count': 1}, factor: 1
#
# The important invariant is that 'names' (the display name) should contain the display
# equivalents to the base units stored in powers (for SI derived units) and lex_names
# (for custom units).
#
# names and lex_names are instances of the NumberDict type.
# NumberDict works like a standard dictionary, except that the keys
# are numeric values, which are used to store the power of each unit
# element.  In addition to normal dictionary access, NumberDicts
# support addition and subtraction.  Adding two NumberDict's merges
# their keys, adding the values of matching keys together.  
#
# There are a few more attributes computed from the above attributes as optimzations:
#
# _name:                The computed name in string form.  For instance, if
#                       names={'km': 1, 'l': -1}, _name will be 'km/l'.  _name
#                       is computed on-demand.
# is_dimensionless:     if powers is all zeros and lex_names is empty, the unit is dimensionless
# is_angle:             If the only non-zero element in powers is 'radians' and lex_names is empty
#
# Unit instances are registered in the module level _unit_table
# dictionary.  This allows you to say Unit('N'), and have it look up
# the right object (including the SI base unit factors).

class Unit(object):
    """Unit of measurement

    A unit is defined by a name (possibly composite), a scaling factor,
    and the exponentials of each of the SI base units that enter into
    it. Units can be multiplied, divided, and raised to rational powers.
    """

    __array_priority__ = 15

    def __new__(cls, *args, **kw):
        """Construct a new unit instance.

        Units can be constructed in several ways:
            one argument: a Unit, WithUnit or str
            three arguments: name, factor, and Unit, WithUnit or str
            otherwise: arguments passed to _init on a new instance
        """
        if len(args) == 1:
            unit = args[0]
            if isinstance(unit, Unit):
                return unit
            elif isinstance(unit, WithUnit):
                return unit.unit
            elif isinstance(unit, str):
                return cls._parse(unit)
            raise Exception('Cannot create Unit for %r' % unit)
        elif len(args) == 3 and isinstance(args[2], (Unit, WithUnit, str)):
            # construct a named unit that is equal to a
            # previously-defined unit, times a factor
            name, factor, unit = args[:3]
            factor = float(factor)
            if isinstance(unit, WithUnit):
                unit, factor = unit.unit, factor * unit._value
            elif isinstance(unit, str):
                unit = cls._parse(unit)
            inst = Unit(name, factor * unit.factor,
                        unit.powers, unit.offset, unit.lex_names)
            return inst
        # construct a new unit
        inst = super(Unit, cls).__new__(cls)
        inst._init(*args, **kw)
        return inst

    def __reduce__(self):
        return (Unit, (self.name,))

    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        return self

    @classmethod
    def _parse(cls, name):
        if name in _unit_table:
            return _unit_table[name]
        unit = _unit_table[''] # start with identity
        try:
            parsed = grammar.unit.parseString(name)
            for sign, list_ in [(1, parsed.posexp), (-1, parsed.negexp)]:
                if isinstance(list_, str):
                    continue
                for elem in list_:
                    num = elem['num'] if 'num' in elem else 1
                    num = -num if 'neg' in elem else num
                    denom = elem['denom'] if 'denom' in elem else 1
                    term = elem['name']
                    if term not in _unit_table:
                        _unit_table[term] = cls._stringUnit(term)
                    unit = unit * _unit_table[term]**(sign*Fraction(num, denom))
        except Exception:
            # TODO handle errors more intelligently here.
            # (might need to change unit grammar)
            # most likely this was a parsing error
            # the manager doesn't guarantee that units
            # will follow the grammar, so not handling this
            # error can crash us and kill our connection.
            # For now, just fall back to a string unit in this case.
            unit = cls._stringUnit(name)
        # if the name of this unit is new, we'll add it to the table
        # of known units.  Otherwise, we'll just return the unit
        # object that is already in the table.  This means units that
        # parse to the same name will end up being the same object.
        return _unit_table.setdefault(unit.name, unit)

    @staticmethod
    def _stringUnit(name):
        """Create a unit that has a name, but is outside the usual SI system."""
        return Unit(NumberDict([(name, 1)]), 1., [0]*9, 0, name)

    def _init(self, names, factor, powers, offset=0, lex_names=''):
        """
        @param names: a dictionary mapping each name component to its
                      associated power (e.g. C{{'m': 1, 's': -1}})
                      for M{m/s}). As a shorthand, a string may be passed
                      which is assigned an implicit power 1.
        @type names: C{dict} or C{str}
        @param factor: a scaling factor
        @type factor: C{float}
        @param powers: the integer powers for each of the nine base units
        @type powers: C{list} of C{int}
        @param offset: an additive offset to the base unit (used only for
                       temperatures)
        @type offset: C{float}
        @param lex_names: a dictionary mapping named components to their
                          associated powers.  lex_names cannot be converted
                          to SI units, but are instead treated as base units
                          themselves.
        @type lex_names: C{dict} or C{str}
        """
        if isinstance(names, str):
            self.names = NumberDict()
            self.names[names] = Fraction(1)
        else:
            self.names = names
        self.factor = factor
        self.offset = offset
        self.powers = [Fraction(p) for p in powers]
        if isinstance(lex_names, str):
            self.lex_names = NumberDict()
            if lex_names:
                self.lex_names[lex_names] = Fraction(1)
        else:
            self.lex_names = lex_names
        self.is_dimensionless = not any(self.powers) and not any(self.lex_names.values())
        self.is_angle = (self.powers[7] == 1 and
                        not any(self.powers[0:7]) and
                        not self.powers[8] and
                        not any(self.lex_names.values()))

    @Lazy
    def name(self):
        num = ''
        denom = ''
        if all(power < 0 for unit, power in self.names.items()):
            # if all powers are negative, use negative exponents
            for unit, power in self.names.items():
                unit += '^' + str(power)
                num += '*' + unit
        else:
            # if some powers are positive, use num/denom
            for unit, power in self.names.items():
                if power != 1 and power != -1:
                    unit += '^' + str(abs(power))
                if power < 0: denom += '/' + unit
                elif power > 0: num += '*' + unit
        num = num[1:] if len(num) else '1' # remove leading '*' from numerator
        name = num + denom
        if name == '1':
            name = ''
        self._name = name
        if name not in _unit_table:
            _unit_table[name] = self
        return name

    @Lazy
    def base_unit(self):
        num = ''
        denom = ''
        for unit, power in (zip(_base_names, self.powers) +
                            self.lex_names.items()):
            if power != 1 and power != -1:
                unit += '^' + str(abs(power))
            if power < 0: denom += '/' + unit
            elif power > 0: num += '*' + unit
        if not len(num):
            num = '1'
        else:
            num = num[1:] # strip leading '*'
        name = num + denom
        if name == '1':
            name = ''
        return Unit(name)

    def __repr__(self):
        return "<Unit '%s'>" % self.name

    def __str__(self):
        return self.name

    def __eq__(self, other):
        if isinstance(other, str):
            try:
                return self.name == other or self == Unit(other)
            except Exception:
                # might fail to convert other to Unit
                pass
        elif isinstance(other, Unit):
            return (self.factor == other.factor and
                    self.offset == other.offset and
                    self.powers == other.powers and
                    self.lex_names == other.lex_names)
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    @cache.lru_cache()
    def _mul_units(self, other):
        if self.offset != 0 or other.offset != 0:
            raise TypeError("cannot multiply units with non-zero offset: '%s', '%s'" % (self, other))
        return Unit(self.names + other.names,
                    self.factor * other.factor,
                    [a + b for a, b in zip(self.powers, other.powers)],
                    self.offset,
                    self.lex_names + other.lex_names)

    def __mul__(self, other):
        if isinstance(other, Unit):
            return self._mul_units(other)
        return WithUnit(other, self)

    __rmul__ = __mul__

    @cache.lru_cache()
    def _div_units(self, other):
        if self.offset != 0 or other.offset != 0:
            raise TypeError("cannot divide units with non-zero offset: '%s', '%s'" % (self, other))
        return Unit(self.names - other.names,
                    self.factor / other.factor,
                    [a - b for a, b in zip(self.powers, other.powers)],
                    self.offset,
                    self.lex_names - other.lex_names)

    def __div__(self, other):
        if isinstance(other, Unit):
            return self._div_units(other)
        return WithUnit(1.0 / other, self)

    @cache.lru_cache()
    def _rdiv_units(self, other):
        if self.offset != 0 or other.offset != 0:
            raise TypeError("cannot divide units with non-zero offset: '%s', '%s'" % (self, other))
        return Unit(other.names - self.names,
                    other.factor / self.factor,
                    [a - b for a, b in zip(other.powers, self.powers)],
                    self.offset,
                    other.lex_names - self.lex_names)

    def __rdiv__(self, other):
        if isinstance(other, Unit):
            return self._rdiv_units(other)
        return WithUnit(other, self**(-1))

    @cache.lru_cache()
    def __pow__(self, other):
        if self.offset != 0:
            raise TypeError("cannot exponentiate unit with non-zero offset: '%s'" % (self,))
        if isinstance(other, int):
            return Unit(self.names * other,
                        pow(self.factor, other),
                        [p * other for p in self.powers],
                        self.offset,
                        self.lex_names * other)
        if isinstance(other, float):
            inv_exp = 1./other
            rounded = int(floor(inv_exp+0.5))
            if abs(inv_exp - rounded) < 1.e-10:
                return self.__pow__(Fraction(1, rounded))
        if isinstance(other, tuple):
            other = Fraction(*other)
        if isinstance(other, Fraction):
            return Unit(other * self.names,
                        pow(self.factor, other),
                        [p * other for p in self.powers],
                        self.offset,
                        other * self.lex_names)
        raise TypeError("Only integer or rational exponents allowed")

    def isCompatible(self, other):     # added 1998/10/01 GPW
        """
        @param other: another unit
        @type other: L{Unit}
        @returns: C{True} if the units are compatible, i.e. if the powers of
                  the base units are the same
        @rtype: C{bool}
        """
        other = Unit(other)
        return self.powers == other.powers and self.lex_names == other.lex_names

    @cache.lru_cache()
    def conversionFactorTo(self, other):
        """
        @param other: another unit
        @type other: L{Unit}
        @returns: the conversion factor from this unit to another unit
        @rtype: C{float}
        @raises TypeError: if the units are not compatible
        """
        other = Unit(other)
        if not self.isCompatible(other):
            raise TypeError("Incompatible units: '%s', '%s'" % (self, other))
        if self.offset != other.offset and self.factor != other.factor:
            raise TypeError(('Unit conversion (%s to %s) cannot be expressed ' +
                             'as a simple multiplicative factor') % (self, other))
        return float(self.factor / other.factor)

    @cache.lru_cache()
    def conversionTupleTo(self, other): # added 1998/09/29 GPW
        """
        @param other: another unit
        @type other: L{Unit}
        @returns: the conversion factor and offset from this unit to
                  another unit
        @rtype: (C{float}, C{float})
        @raises TypeError: if the units are not compatible
        """
        other = Unit(other)
        if not self.isCompatible(other):
            raise TypeError("Incompatible units: '%s', '%s'" % (self, other))

        # let (s1,d1) be the conversion tuple from 'self' to base units
        #   (ie. (x+d1)*s1 converts a value x from 'self' to base units,
        #   and (x/s1)-d1 converts x from base to 'self' units)
        # and (s2,d2) be the conversion tuple from 'other' to base units
        # then we want to compute the conversion tuple (S,D) from
        #   'self' to 'other' such that (x+D)*S converts x from 'self'
        #   units to 'other' units
        # the formula to convert x from 'self' to 'other' units via the
        #   base units is (by definition of the conversion tuples):
        #     ( ((x+d1)*s1) / s2 ) - d2
        #   = ( (x+d1) * s1/s2) - d2
        #   = ( (x+d1) * s1/s2 ) - (d2*s2/s1) * s1/s2
        #   = ( (x+d1) - (d1*s2/s1) ) * s1/s2
        #   = (x + d1 - d2*s2/s1) * s1/s2
        # thus, D = d1 - d2*s2/s1 and S = s1/s2
        factor = float(self.factor / other.factor)
        offset = float(self.offset - (other.offset * other.factor / self.factor))
        return factor, offset

    def isDimensionless(self):
        return self.is_dimensionless

    def isAngle(self):
        return self.is_angle


def convert(*args):
    """Convert between different units.

    There are two calling conventions:
    - 2 args: >>> convert(Value(1, 'mi'), 'm')
    - 3 args: >>> convert(1, 'mi', 'm')
    The string arguments can also be Unit instances.
    """
    if len(args) == 2:
        val, unit = args
        return val.inUnitsOf(unit)
    elif len(args) == 3:
        val, fromUnit, toUnit = args
        factor, offset = Unit(fromUnit).conversionTupleTo(toUnit)
        return (val + offset) * factor
    raise Exception('Must call convert with 2 or 3 args')


_unit_table = {'': Unit(NumberDict(), Fraction(1), [0]*9, 0)}


class WithDimensionlessUnit(object):
    """
    This is a funny class.  It is designed to be subclassed
    along float, complex, or ndarray.  It provides a simplified
    but compatible API as the WithUnit class, but only works
    for dimensionless quantities.  The reason for this is what
    to do with expressions like:

    4. ns * 5. GHz

    Option 1: Return Value(20, '').  This is consistent, but annoying
    because it won't work directly in contexts that expect a float.
    i.e., sin(5*GHz * 2 * np.pi * 4*ns) raises an exception.  It is
    possible to define a __float__() method, but that doesn't catch all
    cases.

    Option 2: return float(20.0).  This is convenient, but makes
    writing generic code harder because the expected methods and
    properties (._value, .inUnitsOf) don't exist.

    Option 3:
    WithDimensionlessUnit(20.0).  This creates a subclass of float
    that has the necessary methods and properties, but in all other
    ways behaves like a float.

    This implements option 3
    """
    __array_priority__ = 15

    __unit = Unit('') # All instances are dimensionless

    def __new__(cls, value):
        obj = super(WithDimensionlessUnit, cls).__new__(cls, value)
        return obj

    def __reduce__(self):
        return (type(self), (self._value,))

    @property
    def unit(self):
        return self.__unit

    @property
    def _value(self):
        return self._numType(self)

    @property
    def units(self):
        return ''

    def __getitem__(self, idx):
        # getitem with a string tries to do unit conversion.  This is not normally
        # particularly useful with dimensionless numbers, but WithUnits(3, '')['mm/m']
        # will give you 3000.0.  If the index isn't a string, pass to the base class
        # implementation
        if isinstance(idx, (str, Unit)):
            return self._value * self.unit.conversionFactorTo(idx)
        else:
            return super(WithDimensionlessUnit, self).__getitem__(idx)

    def inUnitsOf(self, unit):
        if self.unit.conversionFactorTo(unit) != 1.0:
            raise TypeError("Can't convert dimensionless to %s (scale factor must be 1)" % (unit,))
        else:
            return self

    def inBaseUnits(self):
        return self

    def isDimensionless(self):
        return True

    def isCompatible(self, unit):
        return self.unit.isCompatible(unit)

    def sqrt(self):
        return self.__class__(np.sqrt(self._value))

    # These are a whole bunch of operations to make sure that math
    # between WithDimensionlessUnits and float/complex return
    # the right type.  Without this, DimensionlessValues degrade
    # to float as soon as you do math on them.  Not sure if this is
    # actually a problem.
    def __mul__(self, other):
        result = self._value * other
        return WithUnit(result, '')

    __rmul__ = __mul__

    def __add__(self, other):
        result = self._value + other
        return WithUnit(result, '')

    __radd__ = __add__

    def __div__(self, other):
        result = self._value / other
        return WithUnit(result, '')

    __truediv__ = __div__

    def __rdiv__(self, other):
        result = other / self._value
        return WithUnit(result, '')

    __rtruediv__ = __rdiv__

    def __floordiv__(self, other):
        result = self._value // other
        return WithUnit(result, '')

    def __rfloordiv__(self, other):
        result = other // self._value
        return WithUnit(result, '')

    def __sub__(self, other):
        result = self._value - other
        return WithUnit(result, '')

    def __rsub__(self, other):
        result = other - self._value
        return WithUnit(result, '')

    def __neg__(self):
        result = -self._value
        return WithUnit(result, '')

    def __abs__(self):
        result = abs(self._value)
        return WithUnit(result, '')

    # We need to define all the comparison operators this way so that they
    # work when 'other' is an array type.  This is because we have defined
    # an __array_priority__ flag.  When numpy sees this, it assumes we know
    # how to handle numpy arrays and doesn't handle a NotImplemented return
    # value.  The built in float() operators don't know how to handle arrays
    # so we have to override them.
    
    def __lt__(self, other):
        return self._value < other

    def __le__(self, other):
        return self._value <= other

    def __eq__(self, other):
        return self._value == other

    def __ne__(self, other):
        return self._value != other

    def __ge__(self, other):
        return self._value >= other

    def __gt__(self, other):
        return self._value > other
    
    

class DimensionlessFloat(WithDimensionlessUnit, float):
    _numType = float

    def __iter__(self):
        raise TypeError('DimensionlessFloat not iterable')

WithUnit._dimensionlessTypes[float] = DimensionlessFloat
WithUnit._dimensionlessTypes[int] = DimensionlessFloat
WithUnit._dimensionlessTypes[long] = DimensionlessFloat
WithUnit._dimensionlessTypes[np.float64] = DimensionlessFloat
WithUnit._dimensionlessTypes[np.int64] = DimensionlessFloat
WithUnit._dimensionlessTypes[np.float32] = DimensionlessFloat
WithUnit._numericTypes[DimensionlessFloat] = Value


class DimensionlessComplex(WithDimensionlessUnit, complex):
    _numType = complex

    def __iter__(self):
        raise TypeError('DimensionlessComplex not iterable')

WithUnit._dimensionlessTypes[complex] = DimensionlessComplex
WithUnit._dimensionlessTypes[np.complex128] = DimensionlessComplex
WithUnit._numericTypes[DimensionlessComplex] = Complex


class DimensionlessArray(WithDimensionlessUnit, np.ndarray):
    _numType = staticmethod(np.asarray) # The is a 'copy constructor' used in ._value()

    def __new__(cls, value):
        return (np.array(value)*1.0).view(cls)

    def allclose(self, other, *args, **kw):
        return np.allclose(self, other, *args, **kw)

    def __array_wrap__(self, obj):
        '''
        This function is called at the end of uops and similar functions.
        ndarray has some weird logic where reductions like sum() that result
        in zero-rank arrays automatically convert to numpy scalars
        if-and-only-if the type is a ndarray base class.  If we want the same
        we have to do it here
        '''
        if obj.shape == ():
            return WithUnit(obj[()], '')
        else:
            return np.ndarray.__array_wrap__(self, obj)

WithUnit._dimensionlessTypes[np.ndarray] = DimensionlessArray
WithUnit._dimensionlessTypes[list] = DimensionlessArray
WithUnit._numericTypes[DimensionlessArray] = ValueArray


# SI unit definitions
def _addUnit(name, factor, unit, comment='', label='', prefixable=False):
    if name in _unit_table:
        raise KeyError('Unit %s already defined' % name)
    if comment:
        _help.append((name, comment, label, prefixable))
    base = globals()[name] = _unit_table[name] = Unit(name, factor, unit)
    if prefixable:
        for prefix, factor in _prefixes:
            if (prefix + name) in _unit_table:
                raise KeyError('Unit %s already defined' % name)
            unit = Unit(prefix + name, factor, base)
            globals()[prefix + name] = _unit_table[prefix + name] = unit


def addNonSI(name, prefixable=False):
    base = Unit(name, Fraction(1), [0,0,0,0,0,0,0,0,0], lex_names=name)
    base.name # This forces the Lazy evaluation of name, which populates _unit_table
    if prefixable:
        for prefix, factor in _prefixes:
            if (prefix + name) in _unit_table:
                raise KeyError('Unit %s%s already defined' % (prefix, name))
            derived = Unit(prefix+name, factor, base)
            derived.name


_help = []

# SI prefixes
_prefixes = [
    ('Y',  Fraction(10**24)),
    ('Z',  Fraction(10**23)),
    ('E',  Fraction(10**18)),
    ('P',  Fraction(10**15)),
    ('T',  Fraction(10**12)),
    ('G',  Fraction(10**9)),
    ('M',  Fraction(10**6)),
    ('k',  Fraction(10**3)),
    ('h',  Fraction(10**2)),
    ('da', Fraction(10**1)),
    ('d',  Fraction(1, 10**1)),
    ('c',  Fraction(1, 10**2)),
    ('m',  Fraction(1, 10**3)),
    ('u',  Fraction(1, 10**6)),
    ('n',  Fraction(1, 10**9)),
    ('p',  Fraction(1, 10**12)),
    ('f',  Fraction(1, 10**15)),
    ('a',  Fraction(1, 10**18)),
    ('z',  Fraction(1, 10**21)),
    ('y',  Fraction(1, 10**24))
    ]
# SI prefixes
"""
_prefixes = [
    ('Y',  1.e24),
    ('Z',  1.e21),
    ('E',  1.e18),
    ('P',  1.e15),
    ('T',  1.e12),
    ('G',  1.e9),
    ('M',  1.e6),
    ('k',  1.e3),
    ('h',  1.e2),
    ('da', 1.e1),
    ('d',  1.e-1),
    ('c',  1.e-2),
    ('m',  1.e-3),
    ('u',  1.e-6),
    ('n',  1.e-9),
    ('p',  1.e-12),
    ('f',  1.e-15),
    ('a',  1.e-18),
    ('z',  1.e-21),
    ('y',  1.e-24),
    ]
"""
_help.append('SI prefixes:' +
    ', '.join(prefix + ' (%.0E)' % value for prefix, value in _prefixes))


# SI base units
_help.append('SI base units:')

_addUnit('m',   Fraction(1),      [1,0,0,0,0,0,0,0,0], 'meter',     prefixable=True)
_addUnit('g',   Fraction(1,1000), [0,1,0,0,0,0,0,0,0], 'gram',      prefixable=True)
_addUnit('s',   Fraction(1),      [0,0,1,0,0,0,0,0,0], 'second',    prefixable=True)
_addUnit('A',   Fraction(1),      [0,0,0,1,0,0,0,0,0], 'Ampere',    prefixable=True)
_addUnit('K',   Fraction(1),      [0,0,0,0,1,0,0,0,0], 'Kelvin',    prefixable=True)
_addUnit('mol', Fraction(1),      [0,0,0,0,0,1,0,0,0], 'mole',      prefixable=True)
_addUnit('cd',  Fraction(1),      [0,0,0,0,0,0,1,0,0], 'candela',   prefixable=True)
_addUnit('rad', Fraction(1),      [0,0,0,0,0,0,0,1,0], 'radian',    prefixable=True)
_addUnit('sr',  Fraction(1),      [0,0,0,0,0,0,0,0,1], 'steradian', prefixable=True)

_base_names = ['m', 'kg', 's', 'A', 'K', 'mol', 'cd', 'rad', 'sr']


# SI derived units; these automatically get prefixes
_help.append('SI derived units:')

_addUnit('Hz',  Fraction(1),  s**-1,     'Hertz',     '1/s',      prefixable=True)
_addUnit('N',   Fraction(1),  m*kg/s**2, 'Newton',    'm*kg/s^2', prefixable=True)
_addUnit('Pa',  Fraction(1),  N/m**2,    'Pascal',    'N/m^2',    prefixable=True)
_addUnit('J',   Fraction(1),  N*m,       'Joule',     'N*m',      prefixable=True)
_addUnit('W',   Fraction(1),  J/s,       'Watt',      'J/s',      prefixable=True)
_addUnit('C',   Fraction(1),  s*A,       'Coulomb',   's*A',      prefixable=True)
_addUnit('V',   Fraction(1),  W/A,       'Volt',      'W/A',      prefixable=True)
_addUnit('F',   Fraction(1),  C/V,       'Farad',     'C/V',      prefixable=True)
_addUnit('Ohm', Fraction(1),  V/A,       'Ohm',       'V/A',      prefixable=True)
_addUnit('S',   Fraction(1),  A/V,       'Siemens',   'A/V',      prefixable=True)
_addUnit('Wb',  Fraction(1),  V*s,       'Weber',     'V*s',      prefixable=True)
_addUnit('T',   Fraction(1),  Wb/m**2,   'Tesla',     'Wb/m^2',   prefixable=True)
_addUnit('gauss',Fraction(1,10000),T,   'gauss',     '1e-4 T',   prefixable=True)
_addUnit('H',   Fraction(1),  Wb/A,      'Henry',     'Wb/A',     prefixable=True)
_addUnit('lm',  Fraction(1),  cd*sr,     'Lumen',     'cd*sr',    prefixable=True)
_addUnit('lx',  Fraction(1),  lm/m**2,   'Lux',       'lm/m^2',   prefixable=True)
_addUnit('Bq',  Fraction(1),  s**-1,     'Becquerel', '1/s',      prefixable=True)


# Time units
_help.append('Time units:')

_addUnit('min', 60.,     s,   'minute', '60*s')
_addUnit('h',   60.,     min, 'hour',   '60*min')
_addUnit('d',   24.,     h,   'day',    '24*h')
_addUnit('y',   365.25, d,   'year',   '365.25*d')

# Length units
_help.append('Length units:')

_addUnit('inch', 2.54, cm,   'inch', '2.54*cm')
_addUnit('ft',   12.,   inch, 'foot', '12*inch')
_addUnit('mi',   5280, ft,   'mile', '5280*ft')

# Area units
_help.append('Area units:')

_addUnit('acre', 1./640., mi**2, 'acre', '1/640*mi^2')

# Volume units
_help.append('Volume units:')

_addUnit('l',  Fraction(1),  dm**3, 'liter', 'dm^3')
_addUnit('dl', Fraction(1,10),   l, 'deci liter', '0.1 l')
_addUnit('cl', Fraction(1,100),  l, 'centi liter', '0.01 l')
_addUnit('ml', Fraction(1,1000), l, 'milli liter', '0.001 l')

# Force units
_help.append('Force units:')

_addUnit('dyn', 1.e-5, N, 'dyne (cgs unit)', '10^-5 N')

# Energy units
_help.append('Energy units:')

_addUnit('erg', Fraction(1, 10**7), J, 'erg (cgs unit)', '10^-7 J')
_addUnit('cal', 4.184, J, 'calorie', '4.184 J')
_addUnit('kcal', 1000., cal, 'kilocalorie', '1000 cal')
_addUnit('Btu', 1055.05585262, J, 'British thermal unit', '1055.05585262 J')
_addUnit('eV', 1.602176565e-19, C*V, 'electron volt', '1.602176565*10^19 C*V',
         prefixable=True)

# Power units
_help.append('Power units:')

_addUnit('hp', 745.7, W, 'horsepower', '745.7 W')

# Pressure units
_help.append('Pressure units:')

_addUnit('bar', Fraction(10**5), Pa, 'bar (cgs unit)', '10^5 Pa')
_addUnit('atm', 101325., Pa, 'standard atmosphere', '101325 Pa')
_addUnit('torr', 1./760., atm, 'torr = mm of mercury', '1/760 atm')

# Angle units
_help.append('Angle units:')

_addUnit('deg', pi/180, 'rad', 'degrees', 'pi/180 rad')

# Temperature units
_help.append('Temperature units:')

_addUnit('degC', 1.0, Unit({}, 1.0, K.powers, 273.15), 'degrees Celcius')
_addUnit('degF', 1.0, Unit({}, 5./9., K.powers, 459.67), 'degrees Fahrenheit')


# Constants


c       = 299792458.*m/s                   # speed of light
ly      = 1.*c*y                           # light year
mu0     = 4.e-7*pi*N/A**2                  # permeability of vacuum
eps0    = 1/mu0/c**2                       # permittivity of vacuum
G       = 6.67259e-11*m**3/kg/s**2         # gravitational constant
hplanck = 6.62606957e-34*J*s               # Planck constant
hbar    = hplanck/(2*pi)                   # Planck constant / 2pi
e       = 1.60217733e-19*C                 # elementary charge
me      = 9.1093897e-31*kg                 # electron mass
mp      = 1.6726231e-27*kg                 # proton mass
Nav     = 6.0221367e23/mol                 # Avogadro number
k       = 1.380658e-23*J/K                 # Boltzmann constant
wk      = 7*d                              # week
yd      = 3*ft                             # yard
nmi     = 1852.*m                          # Nautical mile
Ang     = 1.e-10*m                         # Angstrom
lyr     = c*y                              # light year
Bohr    = 4*pi*eps0*hbar**2/me/e**2        # Bohr radius
ha      = 10000*m**2                       # hectare
b       = 1.e-28*m**2                      # barn
Hartree = me*e**4/16/pi**2/eps0**2/hbar**2 # Wavenumbers/inverse cm
rootHz  = sqrtHz = Hz**0.5                 # for power spectral density
tsp     = 4.92892159375*ml   # teaspoon
tbsp    = 3*tsp              # tablespoon
floz    = 2*tbsp             # fluid ounce
cup     = 8*floz             # cup
pint    = 16*floz            # pint
qt      = 2*pint             # quart
galUS   = 4*qt               # US gallon
galUK   = 4.54609*l          # British gallon
amu     = 1.6605402e-27*kg   # atomic mass units
oz      = 28.349523125*g     # ounce
lb      = 16*oz              # pound
ton     = 2000*lb            # ton
Ken     = k*K                # Kelvin as energy unit
cali    = 4.1868*J           # international calorie
kcali   = 1000*cali          # international kilocalorie
psi     = 6894.75729317*Pa   # pounds per square inch
degR    = (5./9.)*K          # degrees Rankine
bohr_magneton = 9.2740096820e-24 * J/T # Bohr magneton

_addUnit('phi0', 1.0, hplanck/(2*e), prefixable=True)

# some common textual units (no conversions here)
dB = Unit('dB')
dBm = Unit('dBm')

def description():
    """Return a string describing all available units."""
    s = ''  # collector for description text
    for entry in _help:
        if isinstance(entry, str):
            # headline for new section
            s += '\n' + entry + '\n'
        elif isinstance(entry, tuple):
            name, comment, unit, prefixable = entry
            prefix = '*' if prefixable else ''
            s += '%-8s  %-26s %s\n' % (prefix + name, comment, unit)
        else:
            # impossible
            raise TypeError, 'wrong construction of _help list'
    return s

# add the description of the units to the module's doc string:
__doc__ += '\n' + description()


if __name__ == '__main__':
    # some very basic demonstration code

    l = WithUnit(10., 'm')
    big_l = WithUnit(10., 'km')
    print big_l + l
    t = WithUnit(314159., 's')

    e = 2.7 * Hartree * Nav
    print e.inUnitsOf('kcal/mol')
    print e.inBaseUnits()

    freeze = 0 * Unit('degC')
    print freeze['degF']



from __future__ import absolute_import

from concurrent import futures
from twisted.internet import defer, reactor


def map_future(src, func, *args, **kw):
    """Create a future that applies a function to the result of a source future.

    Args:
        src (concurrent.futures.Future): The future whose result we wish to
            modify.
        func (callable): Function to apply to the result of src. This function
            will be called like func(result, *args, **kw) where result is the
            value produced by src.
        args: Positional args to be passed to func.
        kw: Keyword args to be passed to func.

    Returns:
        A new Future whose result will be the given function applied to the
        result of src. If src fails with an exception, or if func raises an
        exception when called, the error will be propagated to the returned
        future.
    """
    return MappedFuture(src, func, *args, **kw)


class MappedFuture(futures.Future):
    """A Future that receives the result or failure of another Future.

    Args:
        src (concurrent.future.Future): Source future whose value this future
            will receive and possibly modify.
        func (callable | None): Function that will be applied to the result of
            src before setting it on this Future. This function will be called
            like func(result, *args, **kw). If None, the result will be passed
            along unchanged.
        args: Positional args to be passed to func.
        kw: Keyword args to be passed to func.
    """
    def __init__(self, src, func, *args, **kw):
        super(MappedFuture, self).__init__()
        def handle_result(f):
            if not self.set_running_or_notify_cancel():
                return
            if src.cancelled():
                self.set_exception(futures.CancelledError())
                return
            error = src.exception()
            if error is not None:
                self.set_exception(error)
                return
            try:
                result = src.result()
                if func is not None:
                    result = func(result, *args, **kw)
                self.set_result(result)
            except Exception as e:
                self.set_exception(e)
        src.add_done_callback(handle_result)
        self.__src = src

    def cancel(self):
        return super(MappedFuture, self).cancel() and self.__src.cancel()


def call_future(func, *args, **kw):
    """Run func in the twisted reactor thread; return the result as a Future.

    Args:
        func (callable): Function to be run in the twisted reactor thread. May
            return a Deferred.
        *args: Positional arguments to pass to func.
        **kw: Keyword arguments to pass to func.

    Returns:
        (concurrent.future.Future) Future that will receive the result of
        calling func or an exception if it fails.
    """
    f = futures.Future()
    @defer.inlineCallbacks
    def wrapped():
        try:
            result = yield defer.maybeDeferred(func, *args, **kw)
            f.set_result(result)
        except Exception as e:
            f.set_exception(e)
    reactor.callFromThread(wrapped)
    return f


def future_to_deferred(future):
    """Create a twisted Deferred from a Future.

    Args:
        future (concurrent.future.Future): Future whose result or failure we
            wish to capture in a deferred.

    Returns:
        (twisted.internet.defer.Deferred) that will fire when the future
        completes with a result (callback) or error (errback).
    """
    deferred = defer.Deferred()
    def handle_result(future):
        if future.cancelled():
            reactor.callFromThread(deferred.errback, futures.CancelledError())
            return
        error = future.exception()
        if error is not None:
            reactor.callFromThread(deferred.errback, error)
            return
        result = future.result()
        reactor.callFromThread(deferred.callback, result)
    future.add_done_callback(handle_result)
    return deferred




from __future__ import absolute_import

import getpass

from labrad import constants


class Password(object):
    def __init__(self, username='', password=''):
        self.username = username or ''
        self.password = password


# cache of username+password pairs, keyed by host name and port number
_password_cache = {}


def get_password(host=None, port=None, user=None, prompt=True):
    """Get a password from the environment, cache, or prompt.

    To add a password to the cache, use the cache_password function.

    Args:
        host (str or None): The manager host we want to connect to.
        port (int or None): The manager port we want to connect to.
        user (str or None): The user for which to get a password.
        prompt (bool): Whether to prompt the user to enter a password at the
            command line if none is found in the cache or environment vars.

    Returns:
        (Password or None): The password for the given host, port, and user. If
        we did not find a password in the cache or environment vars and prompt
        was False, will return None.
    """
    _init_password_cache()
    if host is None:
        host = constants.MANAGER_HOST
    if port is None:
        port = constants.MANAGER_PORT
    if user is None:
        user = constants.USERNAME or ''
    addr = (host, port)
    if addr in _password_cache and user in _password_cache[addr]:
        return _password_cache[addr][user]
    elif prompt:
        password = _prompt_for_password(host, port, user)
        return Password(user, password)
    else:
        return None


def get_username_and_password(host, port, prompt=True):
    """Get a username and password from the environment, cache, or prompt.

    To add a password to the cache, use the cache_password function.

    Args:
        host (str or None): The manager host we want to connect to.
        port (int or None): The manager port we want to connect to.
        prompt (bool): Whether to prompt the user to enter a password at the
            command line if none is found in the cache or environment vars.

    Returns:
        (Password or None): Credential object containing username and password
        for the given host, port.
    """
    _init_password_cache()
    if host is None:
        host = constants.MANAGER_HOST
    if port is None:
        port = constants.MANAGER_PORT
    addr = (host, port)
    if addr in _password_cache and _password_cache[addr]:
        passwords = _password_cache[addr]
        for user in passwords.keys():
            if user:
                break
        return passwords[user]
    elif prompt:
        user = raw_input('Enter username, or blank for the global user '
                         '({}:{}): '.format(host, port))
        password = _prompt_for_password(host, port, user)
        return Password(user, password)
    else:
        return None


def cache_password(host, port, credential):
    """Add a password to the cache, for the given host and port.

    Cached passwords will be used for subsequent labrad connections to the same
    host and port. See the get_password function.

    Args:
        host (str): The manager host for this password.
        port (int): The manager port for this password.
        credential (Password): The username and password to cache.
    """
    passwords = _password_cache.setdefault((host, port), {})
    passwords[credential.username] = credential


def _prompt_for_password(host, port, user):
    """Prompt user to enter a password for the given host, port, and user."""
    if user:
        msg = 'Enter LabRAD password for {} ({}:{}): '.format(user, host, port)
    else:
        msg = 'Enter LabRAD password ({}:{}): '.format(host, port)
    return getpass.getpass(msg)


def _init_password_cache():
    """Create cache entries for the environment password if needed."""
    if not _password_cache and constants.PASSWORD is not None:
        credential = Password(constants.USERNAME, constants.PASSWORD)
        cache_password(constants.MANAGER_HOST,
                       constants.MANAGER_PORT,
                       credential)
        cache_password(constants.MANAGER_HOST,
                       constants.MANAGER_PORT_TLS,
                       credential)



"""Implements the OAuth 2.0 flow for Mobile and Desktop Apps.

The flow is described in detail in the google developer docs:
https://developers.google.com/identity/protocols/OAuth2InstalledApp

We use OAuth to fetch a token which identifies the user by google account.
This token is then passed to the manager, and the manager in turn calls google
APIs to validate the token and extract the user's email address. It then checks
to see that this email address has been registered as a labrad user.
"""

from __future__ import absolute_import
from __future__ import print_function

import BaseHTTPServer
import json
import os
import threading
import time
import urllib
import urlparse
import webbrowser

from concurrent import futures
import requests


# Page to display after we complete the OAuth flow
_OAUTH_PAGE = """
<html>
<head>
  <title>Labrad authorized!</title>
</head>
<body>
  <h1>Labrad authorized!</h1>
  <p>You may now close this window.</p>
</body>
</html>
"""


OAUTH_URI = 'https://accounts.google.com/o/oauth2/v2/auth'
TOKEN_URI = 'https://www.googleapis.com/oauth2/v4/token'
NO_REDIRECT_URI = 'urn:ietf:wg:oauth:2.0:oob'


TOKEN_PATH = os.path.join(os.path.expanduser('~'), '.labrad', 'client')
TOKEN_FILE = os.path.join(TOKEN_PATH, 'oauth_tokens.json')


# In-memory cache of OAuth tokens. Lasts until the python session is closed.
_token_cache = {}
_token_cache_lock = threading.RLock()


class OAuthToken(object):
    def __init__(self, resp):
        self.access_token = resp['access_token']
        self.expires_at = resp['expires_at']
        self.id_token = resp['id_token']
        self.refresh_token = resp['refresh_token']
        self.token_type = resp['token_type']


def get_token(client_id, client_secret, headless=False, timeout=60):
    """Get id token by asking the user to complete OAuth login flow.

    If we have a cached id_token for this client_id that is not expired, we use
    that. Similarly, if we have a cached refresh_token we will attempt to use
    it to get a new id_token. If refreshing fails, we go through the OAuth flow
    as usual.

    After successfully completing the login flow, we cache the OAuth tokens to
    be used on subsequent login attempts.

    Args:
        client_id (str): The client_id for the app we want to authorize.
        client_secret (str): The client_secret for the app we want to authorize.
        headless (bool): If False, the default, we open a web browser to perform
            the OAuth login and spin up a local http server to receive the
            authorization code. If True, or if we fail to open a webbrowser,
            we print a url for the user to visit to get the authorization code,
            which they must then paste into the terminal. The headless is useful
            in some scenarios such as when tunneling through ssh.
        timeout (int): Amount of time to wait for OAuth login flow to complete,
            in seconds.

    Returns:
        (str) id_token to be passed to the manager to identify the user.
    """
    cached = _get_cached_token(client_id)
    if cached is None:
        print('No cached token. Logging in...')
    else:
        if cached.expires_at > time.time() + 600:
            print('Using cached id_token...')
            return cached
        print('Trying cached refresh_token...')
        try:
            token = _refresh_token(client_id, client_secret,
                                   cached.refresh_token)
            _cache_token(client_id, token)
            return token
        except Exception:
            print('Failed to refresh token. Logging in...')

    def do_headless_login():
        redirect_uri = NO_REDIRECT_URI
        login_uri = _create_login_uri(client_id, redirect_uri)
        print('To obtain an OAuth authorization code, please navigate to the '
              'following URL in a browser:\n\n{}\n'.format(login_uri))
        code = raw_input('When you have completed the login flow, enter the '
                         'code here: ')
        return redirect_uri, code

    if headless:
        redirect_uri, code = do_headless_login()
    else:
        code_future = futures.Future()

        class OAuthHandler(BaseHTTPServer.BaseHTTPRequestHandler):
            def do_GET(self):
                parsed_path = urlparse.urlparse(self.path)
                params = urlparse.parse_qs(parsed_path.query)
                code = params['code'][0]
                code_future.set_result(code)
                self.send_response(200)
                self.send_header('Content-Type', 'text/html')
                self.end_headers()
                self.wfile.write(_OAUTH_PAGE)

            def log_request(self, *args, **kw):
                pass

        # Start local http server to receive redirect on random port
        httpd = BaseHTTPServer.HTTPServer(('localhost', 0), OAuthHandler)
        _, local_port = httpd.server_address

        redirect_uri = 'http://localhost:{}'.format(local_port)
        login_uri = _create_login_uri(client_id, redirect_uri)
        ok = webbrowser.open(login_uri)
        if ok:
            print('Launched web browser to log in with OAuth. Will timeout in '
                  '{} seconds'.format(timeout))
            # Wait to receive code via redirect to local http server
            httpd.timeout = timeout
            httpd.handle_request()
            code = code_future.result(timeout=timeout)
        else:
            redirect_uri, code = do_headless_login()

    data = {
        'code': code,
        'client_id': client_id,
        'client_secret': client_secret,
        'redirect_uri': redirect_uri,
        'grant_type': 'authorization_code'
    }
    response = _send_request(TOKEN_URI, data)
    response['expires_at'] = int(time.time() + response['expires_in'])
    token = OAuthToken(response)
    _cache_token(client_id, token)
    return token


# HTTP helpers.

def _create_login_uri(client_id, redirect_uri):
    """Create URI to start the OAuth flow.

    Args:
        client_id (str): The OAuth client_id of the app to authorize.
        redirect_uri (str): The uri where the user should be redirected to
            receive the authorization code after authorizing the app. This can
            be a localhost url or the special value stored in NO_REDIRECT_URI
            for the "headless" flow.

    Returns:
        (str) URL the user should visit to start the OAuth flow.
    """
    params = {
        'scope': 'openid email profile',
        'redirect_uri': redirect_uri,
        'response_type': 'code',
        'client_id': client_id
    }
    return '{}?{}'.format(OAUTH_URI, urllib.urlencode(params))


def _refresh_token(client_id, client_secret, refresh_token):
    """Attempt to use the given refresh_token to obtain a new id_token.

    Args:
        client_id (str): The OAuth client_id of the app to authorize.
        client_secret (str): The client_secret of the app to authorize.
        refresh_token (str): Token to use to try to fetch a refreshed id_token.

    Returns:
        (OAuthToken) Token containing information from JSON response.
    """
    data = {
        'refresh_token': refresh_token,
        'client_id': client_id,
        'client_secret': client_secret,
        'grant_type': 'refresh_token'
    }
    response = _send_request(TOKEN_URI, data)
    response['expires_at'] = int(time.time() + response['expires_in'])
    response['refresh_token'] = refresh_token
    return OAuthToken(response)


def _send_request(url, params):
    """Helper to send a POST request with the given parameters."""
    r = requests.post(url, data=params)
    if r.status_code != 200:
        raise Exception('request failed: status_code={}'.format(r.status_code))
    return r.json()


# Token cache helpers.

def clear_token_cache():
    """Discard all cached OAuth tokens."""
    with _token_cache_lock:
        _token_cache = {}
        _save_tokens({})


def _get_cached_token(client_id):
    """Get cached token information for the given OAuth client_id.

    Args:
        client_id (str): The OAuth client_id for the app to be authorized.

    Returns:
        (OAuthToken or None): Token for the given client_id or None if we do
        not have a cached token.
    """
    with _token_cache_lock:
        if client_id in _token_cache:
            return _token_cache[client_id]
        tokens = _load_tokens()
        if client_id in tokens:
            return OAuthToken(tokens[client_id])
        return None


def _cache_token(client_id, token, persistent=True):
    """Cache token response for the given client_id.

    Args:
        client_id (str): The OAuth client_id for the app to be authorized.
        resp (OAuthToken): Dict of JSON response parameters from a successful OAuth
            API call. Should contain 'access_token', 'id_token', 'expires_in',
            'refresh_token', and 'token_type' entries. We convert 'expires_in'
            from a relative time to an absolute epoch time and store it as
            'expires_at'.
    """
    with _token_cache_lock:
        _token_cache[client_id] = token
        if persistent:
            tokens = _load_tokens()
            tokens[client_id] = {
                'access_token': token.access_token,
                'expires_at': token.expires_at,
                'id_token': token.id_token,
                'refresh_token': token.refresh_token,
                'token_type': token.token_type
            }
            _save_tokens(tokens)


def _load_tokens():
    """Load dict containing cached tokens, indexed by client_id."""
    _ensure_token_file()
    with open(TOKEN_FILE) as f:
        contents = f.read()
    if not contents:
        return {}
    return json.loads(contents)


def _save_tokens(tokens):
    """Save a dict of cached tokens, indexed by client_id."""
    _ensure_token_file()
    with open(TOKEN_FILE, 'w') as f:
        json.dump(tokens, f, sort_keys=True, indent=4)


def _ensure_token_file():
    """Ensure that the token cache file exists and has appropriate perms."""
    if not os.path.exists(TOKEN_PATH):
        os.makedirs(TOKEN_PATH)
    if not os.path.exists(TOKEN_FILE):
        with open(TOKEN_FILE, 'w') as f:
            pass
    os.chmod(TOKEN_FILE, 0o600)




import time

import pytest

import labrad
from labrad import types as T, util
from labrad.server import LabradServer, setting

class RefreshServer1(LabradServer):
    """First version of the refresh server."""

    name = 'Refresh Test Server'

    @setting(1, returns='s')
    def greet(self, c):
        return 'hello!'

    @setting(2, returns='_')
    def go_away(self, c):
        pass

class RefreshServer2(LabradServer):
    """Second version of the refresh server, with different settings."""
    name = 'Refresh Test Server'

    @setting(1, name='s', returns='s')
    def greet(self, c, name):
        return 'hello, {}!'.format(name)

    @setting(2, a='i', b='i', returns='i')
    def add(self, c, a, b):
        return a + b

class RefreshServer3(LabradServer):
    """Third version of the refresh server, with original settings but differing IDs."""
    name = 'Refresh Test Server'

    @setting(10, returns='s')
    def greet(self, c):
        return 'hello!'

    @setting(20, returns='_')
    def go_away(self, c):
        pass

def test_refresh():
    with labrad.connect() as cxn:

        assert not hasattr(cxn, 'refresh_test_server')

        with util.syncRunServer(RefreshServer1()):
            cxn.refresh()
            assert hasattr(cxn, 'refresh_test_server')
            rts = cxn.refresh_test_server

            def type_list(strs):
                return [T.parseTypeTag(s) for s in strs]

            def same_types(a, b):
                return type_list(a) == type_list(b)

            assert hasattr(rts, 'greet')
            assert same_types(rts.greet.accepts, ['_'])
            assert same_types(rts.greet.returns, ['s'])

            assert hasattr(rts, 'go_away')
            assert same_types(rts.go_away.accepts, ['_'])
            assert same_types(rts.go_away.returns, ['_'])

            assert rts.greet() == 'hello!'

        # pause before refreshing after server disconnect
        time.sleep(1)

        cxn.refresh()
        assert not hasattr(cxn, 'refresh_test_server')

        with util.syncRunServer(RefreshServer2()):
            cxn.refresh()
            assert hasattr(cxn, 'refresh_test_server')

            # check the old rts we got earlier, as well as the cxn attribute
            for srv in [rts, cxn.refresh_test_server]:
                assert hasattr(srv, 'greet')
                assert same_types(srv.greet.accepts, ['s'])
                assert same_types(srv.greet.returns, ['s'])

                assert hasattr(srv, 'add')
                assert same_types(srv.add.accepts, ['(ii)'])
                assert same_types(srv.add.returns, ['i'])

                assert not hasattr(srv, 'go_away')

                assert srv.greet('tester') == 'hello, tester!'
                assert srv.add(100, 101) == 201

        # pause before refreshing after server disconnect
        time.sleep(1)

        cxn.refresh()
        assert not hasattr(cxn, 'refresh_test_server')

        with util.syncRunServer(RefreshServer3()):
            cxn.refresh()
            assert hasattr(cxn, 'refresh_test_server')

            # check the old rts we got earlier, as well as the cxn attribute
            for srv in [rts, cxn.refresh_test_server]:
                assert hasattr(rts, 'greet')
                assert same_types(rts.greet.accepts, ['_'])
                assert same_types(rts.greet.returns, ['s'])

                assert hasattr(rts, 'go_away')
                assert same_types(rts.go_away.accepts, ['_'])
                assert same_types(rts.go_away.returns, ['_'])

                assert rts.greet() == 'hello!'

if __name__ == '__main__':
    pytest.main(['-v', __file__])



"""A collection of miscellaneous support functions and classes.

Most of these used to be located in labrad.util but have been moved here
to break the explicit dependency on twisted.
"""

import collections
import os
import socket
import keyword


def getNodeName():
    return os.environ.get('LABRADNODE', socket.gethostname().lower())


ALLOWED = 'abcdefghijklmnopqrstuvwxyz1234567890_'
FIRST = 'abcdefghijklmnopqrstuvwxyz_'

def mangle(name):
    """Return a modified string that is usable as a method name."""
    newname = ''.join(c if c in ALLOWED else '_' for c in name.lower())
    if newname[0] not in FIRST:
        newname = '_' + newname
    if newname in keyword.kwlist:
        newname = newname + '_'
    return newname

def indent(s, level=1):
    spc = ' ' * (4 * level)
    return '\n'.join(spc + line for line in s.split('\n'))

def extractKey(d, key, default):
    if key not in d:
        return default
    val = d[key]
    del d[key]
    return val

def chunks(s, size):
    '''
    Break a sequence into subsequences of at most size.
    '''
    for i in range(0, len(s), size):
        yield s[i:i+size]

def hexdump(s):
    '''
    Hex dump representation of string.  Each line containes 16 bytes represented
    first as hex codes with printable characters rendered on the right.
    '''
    result = []
    for substr in chunks(s, 16):
        hex_repr = " ".join('%02X' % (ord(x),) for x in substr)
        string_repr = "".join(x if ord(x) > 32 and ord(x) < 127 else '.' for x in substr)
        result.append('%s    %s' % (hex_repr.ljust(47), string_repr))
    return '\n'.join(result)


class SafeIterDict(dict):
    """A dict subclass that allows insertion and deletion while iterating.

    This is accomplished by overriding keys, items, values, etc. to return
    copies of their respective lists, rather than the lists themselves.
    Note that this negates the efficiency gains of using the iter* methods.
    """
    keys = lambda self: list(dict.keys(self))
    items = lambda self: list(dict.items(self))
    values = lambda self: list(dict.values(self))
    iterkeys = lambda self: iter(self.keys())
    iteritems = lambda self: iter(self.items())
    itervalues = lambda self: iter(self.values())
    __iter__ = lambda self: iter(self.keys())

class MultiDict(SafeIterDict):
    """Dictionary with multiple keys to the same value."""
    def __init__(self, *a, **kw):
        dict.__init__(self, *a, **kw)
        self.aliases = {} # mapping from aliases to keys
        self._keys = {} # mapping from keys to aliases

    def __repr__(self):
        items = []
        for k, v in self.iteritems():
            key_str = repr(k)
            if k in self._keys:
                aliases = [key_str] + [repr(a) for a in self._keys[k]]
                key_str = ' or '.join(aliases)
            items.append('%s: %r' % (key_str, v))
        return '{' + ', '.join(items) + '}'

    def __setitem__(self, k, v):
        if isinstance(k, tuple):
            k, aliases = k[0], k[1:]
            for alias in aliases:
                self.aliases[alias] = k
            self._keys[k] = aliases
        dict.__setitem__(self, k, v)

    def __delitem__(self, k):
        if k in self.aliases:
            k = self.aliases[k]
        # delete aliases for this key
        if k in self._keys:
            aliases = self._keys[k]
            for a in aliases:
                # only delete alias if it still points to this key
                if self.aliases[a] == k:
                    del self.aliases[a]
            del self._keys[k]
        dict.__delitem__(self, k)

    def __contains__(self, k):
        if k in self.aliases:
            k = self.aliases[k]
        return dict.__contains__(self, k)

    def __getitem__(self, k):
        if k in self.aliases:
            k = self.aliases[k]
        return dict.__getitem__(self, k)

    def _updateAliases(self, k, *aliases):
        """Update the aliases for a given key."""
        for a in set(self._keys[k]) - set(aliases):
            # remove old aliases, but only if they still
            # point to this key
            if self.aliases[a] == k:
                del self.aliases[a]
        for a in aliases:
            self.aliases[a] = k
        self._keys[k] = aliases

class PrettyMultiDict(MultiDict):
    def __repr__(self):
        return '\n'.join(sorted(self.keys()))


PacketRecord = collections.namedtuple('PacketRecord', ['ID', 'data', 'tag',
                                                       'flat', 'key', 'name'])


FlatPacket = collections.namedtuple('FlatPacket', ['context', 'name',
                                                   'records'])


class PacketResponse(object):
    """Wrapper for response packets from LabRAD servers.

    Attributes are added to access the records in this
    response packet coming from each setting.  For each
    setting called by the packet, we add an attribute to
    access the response from that setting.  If a setting
    is called more than once, all responses from that setting
    are collected into a list.  Responses can be accessed
    by name or ID, unless a key was specified for the call,
    in which case the response is accessible via the key only.
    """
    def __init__(self, resp, server, packet):
        # collect all responses from each setting or key
        temp = collections.defaultdict(list)
        for rec, (ID, data) in zip(packet, resp):
            key = rec.key
            setting = server.settings[ID]
            name, pyName = setting.name, setting._py_name
            # if this record has a key, index by key only
            # otherwise by setting name, python name and ID
            if key is not None:
                name = pyName = ID = key
            keys = (name, pyName, ID)
            temp[keys].append(data)
        # add data for the various settings
        self.settings = MultiDict()
        for (name, pyName, ID), l in temp.items():
            if len(l) == 1:
                l = l[0]
            if isinstance(pyName, str):
                setattr(self, pyName, l)
            self.settings[name, pyName, ID] = l

    def __getitem__(self, key):
        return self.settings[key]



# Copyright (C) 2007  Matthew Neeley
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Super Node

This labrad server finds and controls all available nodes.
"""

from labrad import types as T, util, errors, constants as C
from labrad.devices import DeviceWrapper, DeviceServer
from labrad.server import setting
from labrad.wrappers import connectAsync
from twisted.internet import defer
from twisted.internet.defer import inlineCallbacks, returnValue

class NodeWrapper(DeviceWrapper):
    """A wrapper for a node device."""

    def connect(self, node):
        """Connect to this node."""
        self.node = node # asynchronous wrapper for the node server

    def running_servers(self):
        return self.node.running_servers()

    def available_servers(self):
        return self.node.available_servers()

    def local_servers(self):
        return self.node.local_servers()

class SuperNode(DeviceServer):
    """A server mapping nodes to devices."""
    name = 'Super Node'
    deviceWrapper = NodeWrapper
    testMode = True

    def findDevices(self):
        cxn = self.client
        found = []
        for name in cxn.servers:
            if name.startswith('node_'):
                srv = cxn.servers[name]
                found.append((name, (srv,), {}))
        return found

    # server settings

    @setting(1001, returns=['*s'])
    def running_servers(self, c):
        """Get a list of running servers."""
        dev = self.selectedDevice(c)
        return dev.running_servers()

    @setting(1002, returns=['*s'])
    def available_servers(self, c):
        """Get a list of available servers."""
        dev = self.selectedDevice(c)
        return dev.available_servers()

    @setting(1003, returns=['*s'])
    def local_servers(self, c):
        """Get a list of local servers."""
        dev = self.selectedDevice(c)
        return dev.local_servers()

__server__ = SuperNode()

if __name__ == '__main__':
    from labrad import util
    util.runServer(__server__)



# Copyright (C) 2013 Daniel Sank
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
### BEGIN NODE INFO
[info]
name = GPIB Mock Device Server
version = 1.0
description =

[startup]
cmdline = %PYTHON% %FILE%
timeout = 20

[shutdown]
message = 987654321
timeout = 5
### END NODE INFO
"""

from labrad.server import setting
from labrad.gpib import GPIBManagedServer, GPIBDeviceWrapper
from twisted.internet.defer import inlineCallbacks, returnValue

class Mock0ADevice(GPIBDeviceWrapper):
    @inlineCallbacks
    def initialize(self):
        """Notify that we are being made

        Note the important yield statement
        """
        print("Made a Mock0A device")
        yield

class Mock1BDevice(GPIBDeviceWrapper):
    @inlineCallbacks
    def initialize(self):
        """Notify that we are being made

        Note the important yield statement
        """
        print("Made a Mock1B device")
        yield

class GpibMockDeviceServer(GPIBManagedServer):
    """Provides basic CW control for Anritsu 68367C Microwave Generators"""
    name = 'GPIB Mock Device Server'
    deviceWrappers = {
                      'MOCK 0A': Mock0ADevice,
                      'MOCK 1B': Mock1BDevice
                      }

__server__ = GpibMockDeviceServer()

if __name__ == '__main__':
    from labrad import util
    util.runServer(__server__)



import pytest

from labrad import crypto


def test_fingerprint():
    """Verify that fingerprint matches what we get from openssl.

    To get the fingerprint of a certificate with openssl, you can do:
    $ openssl x509 -fingerprint -in <certfile>
    """
    cert = """
    -----BEGIN CERTIFICATE-----
    MIICqjCCAZKgAwIBAgIILDVF+J1HIWowDQYJKoZIhvcNAQELBQAwFDESMBAGA1UE
    AwwJbG9jYWxob3N0MCAXDTE0MDgyODE3MTUyM1oYDzk5OTkxMjMxMjM1OTU5WjAU
    MRIwEAYDVQQDDAlsb2NhbGhvc3QwggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEK
    AoIBAQCVPL4WfLACl/oR+RNolkwxfyxcj8IdTrzrUXmY1zBBncdHhHOPijJZHdP5
    vmD4HO2kb793V4uT02gnMq+MiYr8fNui7yRYnInG3GsgwJKQTH/5objHN6JZl505
    82r/aas0rDjRRc25WXGSCuPuW+h1u0f2N4PuBwN6iCmv0FDhEz7C3FgDWDSk4heY
    xLglX2wnkrD4nob4NXWqX7VZ5A9KYBKQPhpOczg+vtv1nd3CnU5JN9GbC3AZ+4wF
    Zb0xMpLDgn4T5z0TjOhJOwCbDOSsJWdC0q6UQe4ln5GubzKjPaJWVR4VXOsVuqoU
    q4fLed5vHSsQUc0C6qElvak0xULbAgMBAAEwDQYJKoZIhvcNAQELBQADggEBACVZ
    H5UJfpvujJ5tV0Mj9Jho0f5Wf/KARTa/5NL6lbM3KPSrrOZGdnH2Hag6me3JMJ+y
    kpxCc5HQnhF+Ep21Y5tFo0Ex2/FRGSfxZZVL0ERjMYnJpzjbnB4S5VPYW1LB+ctL
    +kwNc+sc8up986zNZnzxRY5hllvmC82Bn24dCVECzy3fgczVpOSh4pLeF+sXYOA1
    2ZT081GtWsEjebCndRoTtInTkqdtsSLHvznAi8YQ7lhtow/sAr5hbUaGwdrROUaq
    6Z+dzh7LHrTmfqmefNGGqi+hWKSU+fxGYQ+QDUjCD5J1dfnnueeHgYIeiQ9ZjnyE
    f3brW92sWUeqydhqD5A=
    -----END CERTIFICATE-----
    """
    expected_fingerprint = "5E:DF:14:6E:87:0F:2D:15:FD:B2:07:4E:6B:43:9A:21:84:58:F7:70"
    assert crypto.fingerprint(cert) == expected_fingerprint


if __name__ == '__main__':
    pytest.main(['-v', __file__])
